/* Driver for RATOC SMA03 SmartMedia reader
 *
 * sma03 driver v0.22
 *
 * Beta release
 *
 * Current development and maintenance by:
 *   (c) 2001 RATOC Systems, Inc.
 *
 * The RATOC SMA03 SmartMedia reader/writer uses the FUJITSU MB89589 chip.
 * This driver translates the "real" SCSI commands to the SMA03 vendor
 * unique commands.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "transport.h"
#include "protocol.h"
#include "usb.h"

// #define CONFIG_USB_STORAGE_DEBUG
#include "debug.h"
#include "sma03.h"

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/slab.h>

#define short_pack(lsb,msb) ( ((u16)(lsb)) | ( ((u16)(msb))<<8 ) )
#define short_pack2(p) ( ((u16)(p[1])) | ( ((u16)(p[0]))<<8 ) )
#define LSB_of(s) ((s)&0xFF)
#define MSB_of(s) ((s)>>8)

#define SMA_GOOD        0
#define SMA_ERROR       -1
#define SMA_HALF_OK     -2
#define SMA_TIMEOUT     (-ETIMEDOUT)  /* -110 */

#define SMA_ECC_MARKED_F0  -1
#define SMA_ERROR_TRANSFER_WRITE_BOMB   -(0x63)
#define SMA_WRITE_PROTECTED     -3
#define SMA_MEDIA_ERROR         -4
#define SMA_DEVICE_ERROR        -5
#define SMA_NOT_EXECUTE         -6
#define SMA_VERIFY_CHECK_ERROR  -7


#define EX_HILO_WORD(word)   ( ((unsigned short)word >> 8) | (((unsigned short)word & 0x00ff) << 8) )

/* for debug */
#define DebugDump(PTR, LEN)  { int i; for (i = 0; i<LEN; i++) printk("%02X ", PTR[i]); }

/*
 * Buffer of Request Sense Data
 */
static unsigned char *current_sense_data = NULL;
static unsigned char noSense[] = {
    0x70, 0x00,                          /* 00-01 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00  /* 02-07 */
};
static unsigned char mediumNotPresent[] = {
    0x70, 0x00,                                                    /* 00-01 */
    NOT_READY, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x00, 0x00, 0x00, 0x00,  /* 02-11 */
    0x3A, 0x00, 0x00, 0x00, 0x00, 0x00                            /* 12-17 */
};
static unsigned char notReady[] = {
    0x70, 0x00,
    UNIT_ATTENTION, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x00, 0x00, 0x00, 0x00,
    0x28, 0x00, 0x00, 0x00, 0x00, 0x00
};
static unsigned char illegalRequest[] = {
    0x70, 0x00,
    0x05, 0x00, 0x00, 0x00, 0x00, 0x0A, 0x00, 0x00, 0x00, 0x00,
    0x22, 0x00, 0x00, 0x00, 0x00, 0x00
};

static int sma03_ResultSave;
/*
 * Send a control message and wait for the response.
 *
 * us - the pointer to the us_data structure for the device to use
 *
 * request - the URB Setup Packet's first 6 bytes. The first byte always
 *  corresponds to the request type, and the second byte always corresponds
 *  to the request.  The other 4 bytes do not correspond to value and index,
 *  since they are used in a custom way by the SCM protocol.
 *
 * xfer_data - a buffer from which to get, or to which to store, any data
 *  that gets send or received, respectively, with the URB. Even though
 *  it looks like we allocate a buffer in this code for the data, xfer_data
 *  must contain enough allocated space.
 *
 * xfer_len - the number of bytes to send or receive with the URB.
 *
 */

static int sma03_send_control(struct us_data *us,
                int pipe,
                unsigned char request,
                unsigned char requesttype,
                unsigned short value,
                unsigned short index,
                unsigned char *xfer_data,
                unsigned int xfer_len) {

    int result;

    result = usb_stor_control_msg(us, pipe,
                        request, requesttype, value, index,
                        xfer_data, xfer_len);


    if (result < 0) {
        sma03_ResultSave = result;
        /* if the command was aborted, indicate that */
        if (result == -ENOENT)
            return USB_STOR_TRANSPORT_ABORTED;

        /* a stall is a fatal condition from the device */
        if (result == -EPIPE) {
            US_DEBUGP("-- Stall on control pipe. Clearing\n");
            result = usb_clear_halt(us->pusb_dev, pipe);
            US_DEBUGP("-- usb_clear_halt() returns %d\n", result);
            return USB_STOR_TRANSPORT_FAILED;
        }

        /* Uh oh... serious problem here */
        return USB_STOR_TRANSPORT_ERROR;
    }

    return USB_STOR_TRANSPORT_GOOD;
}
/*
 *  Transfer Bulk data
 */
static int
sma03_raw_bulk(struct us_data *us,
                int direction,
                unsigned char *data,
                unsigned int len )
{
    int result;
    int act_len;
    int pipe = 0;

    switch (direction) {
    case SMA03_DATA_WRITE:
        pipe = usb_sndbulkpipe(us->pusb_dev, 1);   /* EP1 */
        break;
    case SMA03_DATA_READ:
        pipe = usb_rcvbulkpipe(us->pusb_dev, 2);   /* EP2 */
        break;
    case SMA03_CTRL_WRITE:
        pipe = usb_sndbulkpipe(us->pusb_dev, 3);   /* EP3 */
        break;
    default:
        printk("sma03_raw_bulk: Invalid direction: %d\n", direction);
        return US_BULK_TRANSFER_FAILED;
    }

    result = usb_stor_bulk_msg(us, data, pipe, len, &act_len);
   if (len != act_len) {
    printk("sma03_raw_bulk: direction: %d usb_stor_bulkmsg:\n", direction);
    printk(" pipe: %04X result: %d\n", pipe, result);
    printk(" len: %04X act_len: %04X\n", len, act_len);
   }

    /* if we stall, we need to clear it before we go on */
    if (result == -EPIPE) {
        US_DEBUGP("EPIPE: clearing endpoint halt for"
                   " pipe 0x%x, stalled at %d bytes\n",
                   pipe, act_len);
        usb_clear_halt(us->pusb_dev, pipe);
    }

    if (result) {
        sma03_ResultSave = result;

        /* NAK - that means we've retried a few times already */
        if (result == -ETIMEDOUT) {
            US_DEBUGP("usbat_raw_bulk(): device NAKed\n");
            return US_BULK_TRANSFER_FAILED;
        }

        /* -ENOENT -- we canceled this transfer */
        if (result == -ENOENT) {
            US_DEBUGP("usbat_raw_bulk(): transfer aborted\n");
            return US_BULK_TRANSFER_ABORTED;
        }

        if (result == -EPIPE) {
            US_DEBUGP("usbat_raw_bulk(): output pipe stalled\n");
            return USB_STOR_TRANSPORT_FAILED;
        }

        /* the catch-all case */
        US_DEBUGP("us_transfer_partial(): unknown error\n");
        return US_BULK_TRANSFER_FAILED;
    }

    if (act_len != len) {
        US_DEBUGP("Warning: Transferred only %d bytes\n", act_len);
        return US_BULK_TRANSFER_SHORT;
    }

    US_DEBUGP("Transferred %d of %d bytes\n", act_len, len);

    return US_BULK_TRANSFER_GOOD;
}  /* sma03_raw_bulk */

/*
 * Note: direction must be set if command_len == 0.
 */

static int
sma03_bulk_transport(struct us_data *us,
                          int direction,
                          unsigned char *data,
                          unsigned int len,
                          int use_sg)
{

    int result = USB_STOR_TRANSPORT_GOOD;
    int transferred = 0;
    int a_len = 0;
    int i;
    struct scatterlist *sg;
    char string[64];

    if (len==0)
        return USB_STOR_TRANSPORT_GOOD;

        /* transfer the data */

    if (direction != SMA03_DATA_READ) {

        /* Debug-print the first 48 bytes of the write transfer */

        if (!use_sg) {
            strcpy(string, "wr: ");
            for (i=0; i<len && i<48; i++) {
                sprintf(string+strlen(string), "%02X ", data[i]);
                if ((i%16)==15) {
                    US_DEBUGP("%s\n", string);
                    strcpy(string, "wr: ");
                }
            }
            if ((i%16)!=0)
                US_DEBUGP("%s\n", string);
        }
    }


    US_DEBUGP("data %s transfer %d sg buffers %d\n",
              ( direction==SMA03_DATA_READ ? "in" : "out"),
              len, use_sg);

    if (!use_sg)
        result = sma03_raw_bulk(us, direction, data, len);
    else {
        sg = (struct scatterlist *)data;
        for (i=0; i<use_sg && transferred<len; i++) {
            a_len = (len-transferred) > sg[i].length ?
                    sg[i].length : len-transferred;
            if (a_len == 0)
                break;
            result = sma03_raw_bulk(us, direction,
                sg[i].address, a_len);
            if (result!=US_BULK_TRANSFER_GOOD)
                break;
            transferred += sg[i].length;
        }  /* for (i = 0...use_sg */
    }

    if (direction == SMA03_DATA_READ) {

    /* Debug-print the first 48 bytes of the read transfer */

        if (!use_sg) {
            strcpy(string, "rd: ");
            for (i=0; i<len && i<48; i++) {
                sprintf(string+strlen(string), "%02X ", data[i]);
                if ((i%16)==15) {
                    US_DEBUGP("%s\n", string);
                    strcpy(string, "rd: ");
                }
            }
           if ((i%16)!=0)
               US_DEBUGP("%s\n", string);
        }
    }

    return result;
}


static int
sma03_send_bulkcontrol(struct us_data *us,
                int pipe,
                unsigned char request,
                unsigned char requesttype,
                union CD_ADDR address,
                unsigned short count ) {

    int result;
    unsigned short value;
    unsigned short index;
    unsigned char  ctrlmsg[8];

    value = (unsigned short)(address.b.adr2 << 8 | address.b.adr1);
    index = (unsigned short)(address.b.adr0 << 8 | 0x00);

    ctrlmsg[0] = requesttype;
    ctrlmsg[1] = request;
    ctrlmsg[2] = address.b.adr1;   /* PA8-15 */
    ctrlmsg[3] = address.b.adr2;   /* PA16-23 */
    ctrlmsg[4] = 0;                /* CA0-7 */
    ctrlmsg[5] = address.b.adr0;   /* PA0-7 */
    ctrlmsg[6] = count & 0x00ff;
    ctrlmsg[7] = (count & 0xff00) >> 8;

//  printk("ctrlmsg:");   DebugDump(ctrlmsg, 16);  printk("\n");

    result = usb_stor_control_msg(us, pipe,
                        request, requesttype, value, index,
                        ctrlmsg, 8);

    if (result < 0) {
        sma03_ResultSave = result;
        /* if the command was aborted, indicate that */
        if (result == -ENOENT)
            return USB_STOR_TRANSPORT_ABORTED;

        /* a stall is a fatal condition from the device */
        if (result == -EPIPE) {
            US_DEBUGP("-- Stall on control pipe. Clearing\n");
            result = usb_clear_halt(us->pusb_dev, pipe);
            US_DEBUGP("-- usb_clear_halt() returns %d\n", result);
            return USB_STOR_TRANSPORT_FAILED;
        }

        /* Uh oh... serious problem here */
        return USB_STOR_TRANSPORT_ERROR;
    }

    return USB_STOR_TRANSPORT_GOOD;
}  /* sma03_send_bulkcontrol */

/*
 * =================================================================
 */
/*
 *  Reset Media
 */
static int
sma03_StanbyMode(struct us_data *us)
{
    return ( sma03_send_control(us, usb_sndctrlpipe(us->pusb_dev, 0),
                SMA03_OUT_DATA,
                USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                SMA03_MODE_STANDBY, SMA03PORT_USB_MODE,
                NULL, 0) );
}

/*
 *  Reset Media
 */
static int
sma03_ResetMedia(struct us_data *us)
{
    return sma03_send_control(us, usb_sndctrlpipe(us->pusb_dev, 0),
                SMA03_RESET_MEDIA,
                USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, 0,
                NULL, 0);
}

/*
 *
 */
static int
sma03_GetFwInfo( struct us_data *us )
{
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    memset( (unsigned char *)&(info->FWInfo), 0x00, 8);
    return ( sma03_send_control(us, usb_rcvctrlpipe(us->pusb_dev, 0),
                SMA03_GET_FW_INFO,
                USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, 0,
                (unsigned char *)&(info->FWInfo),
                8) );
}

/*
 *  Read Data and Redundant of one page  (256bytes or 512bytes)
 */
static int
sma03_ReadSinglePage( struct us_data *us,
                union CD_ADDR address,
                unsigned char *data_buf,
                unsigned char *redu_buf)
{

    int     result;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    current_sense_data = NULL;
    if ((result = sma03_send_bulkcontrol(us,
                    usb_sndbulkpipe(us->pusb_dev, 3),
                    SMA03_READ_PAGE,
                    USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                    address,
                    0)) != USB_STOR_TRANSPORT_GOOD)  {
        US_DEBUGP("sma03_ReadSinglePage: sma03_send_control result: %d\n",
                 sma03_ResultSave);
        return result;
    }
    /* Read data */
    if ((result = sma03_bulk_transport(us, SMA03_DATA_READ,
                data_buf, info->pagesize,
                0)) != USB_STOR_TRANSPORT_GOOD)  {
        return result;
    }

    /* Read redu */
    if ((result = sma03_bulk_transport(us, SMA03_DATA_READ,
                redu_buf, info->redusize,
                0)) != USB_STOR_TRANSPORT_GOOD)  {
        return result;
    }

    return USB_STOR_TRANSPORT_GOOD;
}  /* sma03_ReadSinglePage */

/*
 *  Read Redundant of one page  (8bytes or 16bytes)
 */
static int
sma03_ReadSingleRedu( struct us_data *us,
                union CD_ADDR address,
                unsigned char *redu_buf)
{

    int     result;
    unsigned short value;
    unsigned short index;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

     /* PA8-15, PA16-23 */
    value = (unsigned short)(address.b.adr2 << 8 | address.b.adr1);
     /* PA0-7 */
    index = (unsigned short)(address.b.adr0 << 8 | 0x00);
    US_DEBUGP("sma03_ReadSingleRedu: address: %08lX  content: %p"
           "  value: %04X  index: %04X\n",
           address.Adrs, redu_buf, value, index);
    current_sense_data = NULL;
    if ((result = sma03_send_control(us,
                    usb_rcvctrlpipe(us->pusb_dev, 3),
                    SMA03_READ_REDU,
                    USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                    value, index,
                    NULL, 0)) != USB_STOR_TRANSPORT_GOOD)  {
        US_DEBUGP("sma03_ReadSingleRedu: sma03_send_control result: %d\n",
                        sma03_ResultSave);
        return result;
    }

    /* Read redu */
    if ((result = sma03_bulk_transport(us, SMA03_DATA_READ,
                redu_buf, info->redusize,
                0)) != USB_STOR_TRANSPORT_GOOD)  {
        return result;
    }
/* { int i; for (i = 0; i<16; i++) printk("%02X ", redu_buf[i]); } */

    return USB_STOR_TRANSPORT_GOOD;
}  /* sma03_ReadSingleRedu */

/*
 *  Read Data and Redundant of two pages
 */
static int
sma03_ReadDoublePages( struct us_data *us,
                union CD_ADDR address,
                unsigned char *data_buf,
                unsigned char *redu_buf)
{

    int     result;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    current_sense_data = NULL;

    if ( (result = sma03_ReadSinglePage( us, address,
                                data_buf, redu_buf) ) )
        return result;

    return ( sma03_ReadSinglePage( us,
                                (union CD_ADDR)(address.Adrs + 1),
                                data_buf + info->pagesize,
                                redu_buf + info->redusize) );
}  /* sma03_ReadDoublePages */

/*
 *  Read a block (use READ_PAGE command)
 */
static int sma03_ReadPages( struct us_data *us,
                unsigned long address,
                unsigned short pages,
                unsigned char *content,
                int use_sg) {

    int     result = USB_STOR_TRANSPORT_GOOD;
    int     i;
    struct scatterlist *sg = NULL;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    US_DEBUGP("sma03_ReadPages: pages=%d\n", pages);
    if (pages == 1) {
        return sma03_ReadSinglePage( us, (union CD_ADDR)address, content, info->rdt);
    }

    sg = (struct scatterlist *)content;
    for (i = 0; i < pages; i++, address++) {
        printk("[%ld/%p]",address, sg[i].address);
/*
            result = sma03_ReadSinglePage( us, address, sg[i],address, info->rdt);
            if (result != USB_STOR_TRANSPORT_GOOD)  {
                return result;
            }
 */
    }
    return result;
}

/*
 *
 */
static int
sma03_ReadBlockPages( struct us_data *us,
                union CD_ADDR address,
                unsigned char  *buf, /* PBLOCKREAD_MEMORYMAP512 */
                unsigned short pages)
{

    int     result;
    int     loop;
    int     tfrlen;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    if (info->pagesize == Page256M)
        tfrlen = sizeof(BLOCKREAD_MEMORYMAP256)
                     * (pages / info->PagePerSector);
    else
        tfrlen = sizeof(BLOCKREAD_MEMORYMAP512)
                     * (pages / info->PagePerSector);

//  printk("sma03_ReadBlockPages: addres: %08lX \n", address.Adrs);
//  printk(" tfrlen: %08X pages: %04X\n", tfrlen, pages);

    for (loop = 0; loop < 5; loop++) {
        if ( (result = sma03_send_bulkcontrol(us,
                usb_sndbulkpipe(us->pusb_dev, 3), /* Control */
                SMA03_READ_BLOCK,
                USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                address, pages)
             ) != USB_STOR_TRANSPORT_GOOD)  {
            US_DEBUGP("sma03_ReadBlockPages: "
                "sma03_send_bulkcontrol: %d retry: %d\n",
                sma03_ResultSave, loop +1 );
            continue;
        }

        if ( (result = sma03_bulk_transport(us,
                SMA03_DATA_READ,
                buf, tfrlen, 0)
              ) == USB_STOR_TRANSPORT_GOOD)  {
            break;
        }
        US_DEBUGP("sma03_ReadBlockPages: "
               "sma03_bulk_transport: %d retry: %d\n",
               sma03_ResultSave, loop +1 );
        sma03_ResetMedia(us);   /* if error then reset Media */
    }

    return result;
}  /* sma03_ReadBlockPages */

/*
 *  Read Sectors (use READ_PAGE command)
 */
static int
sma03_ReadSectors( struct us_data *us,
                PBLOCKREAD_MEMORYMAP512 mapptr512,
                unsigned short Sectors ) {

    int  result;
    int  loop;
    unsigned int  sector;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    for (sector = 0; sector < Sectors; sector++) {
        for (loop = 0; loop < 5; loop++ ) {
            if ( ( result = (info->read_sector)( us,
                         info->Card, mapptr512->DataArea,
                         (unsigned char *)(&mapptr512->RedundantArea) ) )
                         ) {
                sma03_ResetMedia(us);   /* if error then reset Media */
                return result;
            }
        }
        info->Card.Adrs += info->PagePerSector;
        mapptr512 ++;
    }

    return USB_STOR_TRANSPORT_GOOD;
}  /* sma03_ReadSectors */

/*
 *  Read Sectors (use READ_PAGE command)
 */
static int
sma03_ReadReduSectors( struct us_data *us,
                unsigned char  *redu_buf,
                unsigned short Sectors ) {

    int  result = 0;
    int  loop;
    unsigned int  sector;
    unsigned char  *pbuf;
    unsigned char  *dumy_buf;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    if ( (dumy_buf = kmalloc(SMA_SECTOR, GFP_KERNEL)) == NULL)
            return USB_STOR_TRANSPORT_ERROR;

    pbuf = redu_buf;

    for (sector = 0; sector < Sectors; sector++) {
        for (loop = 0; loop < 5; loop++ ) {
            if ( ( result = (info->read_sector)( us,
                         info->Card,
                         dumy_buf, pbuf ) )
                          == USB_STOR_TRANSPORT_GOOD)
                break;    /* if success then out of loop */
            sma03_ResetMedia(us);  /* if error then reset Media */
            /* Retry */
            printk("sma03_ReadReduSectors: Page=%ld Retry:%d\n",
            info->Card.Adrs, loop + 1 );
        }
        if (result) {
    printk("sma03_ReadReduSectors: Sectors=%d Page=%ld Error:%d\n",
            Sectors, info->Card.Adrs, result);
            kfree(dumy_buf);
            return result;
        }
        info->Card.Adrs += info->PagePerSector;
        pbuf += REDUD;
    }
    kfree(dumy_buf);

    return USB_STOR_TRANSPORT_GOOD;
}  /* sma03_ReadReduSectors */


static int
sma03_BlockRead( struct us_data *us,
        PBLOCKREAD_MEMORYMAP512 MediaBufPtr,
        unsigned short sectors )
{
    int            bret = 0;
    int            loop;
    unsigned short pages;
    unsigned char  *ptr_dest = (unsigned char *)MediaBufPtr;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;


    pages = sectors * info->PagePerSector;

    if ( info->pagesize == Page512M ) { // 512Byte/Media
        for ( loop = 0; loop < 5; loop++ ) {
//            bret = sma03_ReadSectors(us, ptr_dest, sectors);

            bret = sma03_ReadBlockPages( us,
                info->Card, ptr_dest, pages);
            if ( bret == USB_STOR_TRANSPORT_GOOD ) {
                break;
            } else {
                printk("\n\n DANGER!!!*** \n\n"
                        "*** sma03_BlockRead  USB TRANSFER Fail!! %02x\n\n",
                        bret );
                sma03_ResetMedia(us);   /* if error then reset Media */
            }
        }
        if ( bret ) {
            return bret;
        }
        return USB_STOR_TRANSPORT_GOOD;
    }

    /*
     * 256Byte/Media
     */

  {
    int      iii;
    PBLOCKREAD_MEMORYMAP512 mapptr512 = (PBLOCKREAD_MEMORYMAP512)ptr_dest;
    PBLOCKREAD_MEMORYMAP256 mapptr256 = (PBLOCKREAD_MEMORYMAP256)ptr_dest;

    for ( loop = 0; loop < 5; loop++ ) {
        bret = sma03_ReadBlockPages( us,
            info->Card, ptr_dest, pages);
        if ( bret == USB_STOR_TRANSPORT_GOOD ) {
            break;
        } else {
            printk("\n\n DANGER!!!*** \n\n"
                    "*** sma03_BlockRead  USB TRANSFER Fail!! %02x\n\n",
                    bret );
            sma03_ResetMedia(us);   /* if error then reset Media */
        }
    }
    if ( bret ) {
        return bret;
    }

    for ( iii=0; iii < sectors; iii++ ) {
        /*
         *  [  Data0(256) + Redund0(8) + Padding0(56)
         *   + Data1(256) + Redund1(8) + Padding1(56)]
         *  ==>  [  ... Data1(256) + Reserved(8) + Redund1(8) ... ]
         */
        memcpy( (unsigned char *)&(mapptr256->RedundantArea1) + 8,
                    (unsigned char *)(&(mapptr256->RedundantArea1)), 8 );
        /*
         *  [  Data0(256) + Redund0(8) + Padding0(56)
         *   + Data1(256) + Reserved(8) + Redund1(8) + Padding1(48)]
         *  ==>  [  ... Data1(256) + Redund0(8) + Redund1(8) ... ]
         */
        memcpy( (unsigned char *)(&(mapptr256->RedundantArea1)),
                    (unsigned char *)(&(mapptr256->RedundantArea0)), 8 );
        /*
         *  mapptr256 [ Data0(256) + Data1(256) + Redund0(8) + Redund1(8) ]
         *  ==> mapptr512 [ Data0(256) ]
         *  ==> mapptr512 [ Data0(266) + Data1(256) + Redund0(8) + Redund1(8) ]
         */
        memcpy( mapptr512->DataArea, mapptr256->DataArea0, 256 );
        memcpy( mapptr512->DataArea + 256,
                        (unsigned char *)(mapptr256->DataArea1), 256+16 );

        mapptr256++;        // +=640
        mapptr512++;        //+=576
    }
          // mod_d
  }

    return USB_STOR_TRANSPORT_GOOD;

}  /* sma03_BlockRead */


/*
 *  Write Data and Redundant of one page  (256bytes or 512bytes)
 */
static int
sma03_WriteSinglePage( struct us_data *us,
                union CD_ADDR address,
                unsigned char *data_buf,
                unsigned char *redu_buf)
{

    int     result;
    int     loop;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    current_sense_data = NULL;
    for (loop = 0; loop < 5; loop++) {
        if ( (result = sma03_send_bulkcontrol(us,
                usb_sndbulkpipe(us->pusb_dev, 3), /* Control */
                SMA03_WRITE_PAGE,
                USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                address, 0)
              ) != USB_STOR_TRANSPORT_GOOD)  {
            printk("sma03_WriteSinglePage: "
                "sma03_send_bulkcontrol: %d retry: %d\n",
                sma03_ResultSave, loop +1 );
            continue;
        }

        /* Write data */
        if ((result = sma03_bulk_transport(us,
                        SMA03_DATA_WRITE,
                        data_buf, info->pagesize,
                        0)) != USB_STOR_TRANSPORT_GOOD)  {
            continue;
        }

        /* Write redu */
        if ((result = sma03_bulk_transport(us,
                        SMA03_DATA_WRITE,
                        redu_buf, info->redusize,
                        0)) == USB_STOR_TRANSPORT_GOOD)  {
            break;
        }

    }

    return result;

}  /* sma03_WriteSinglePage */

/*
 *  Write Data and Redundant of two pages
 */
static int
sma03_WriteDoublePages( struct us_data *us,
                union CD_ADDR address,
                unsigned char *data_buf,
                unsigned char *redu_buf)
{

    int     result;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    printk("sma03_WriteDoublePages: address: %08lX  data_buf: %p\n",
           address.Adrs, data_buf);
    current_sense_data = NULL;

    if ( (result = sma03_WriteSinglePage( us, address,
                                data_buf, redu_buf) ) )
        return result;

    return ( sma03_WriteSinglePage( us,
                                (union CD_ADDR)(address.Adrs + 1),
                                data_buf + info->pagesize,
                                redu_buf + info->redusize) );
}  /* sma03_WriteDoublePages */

/*
 *  Write a block (use READ_PAGE command)
 */
static int sma03_WritePages( struct us_data *us,
                unsigned long address,
                unsigned short pages,
                unsigned char *content,
                int use_sg) {

    int     result = USB_STOR_TRANSPORT_GOOD;
    int     i;
    struct scatterlist *sg = NULL;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    printk("sma03_WritePages: pages=%d\n", pages);
    if (pages == 1) {
        return sma03_WriteSinglePage( us, (union CD_ADDR)address, content, info->rdt);
    }

    sg = (struct scatterlist *)content;
    for (i = 0; i < pages; i++, address++) {
        printk("[%ld/%p]",address, sg[i].address);
/*
            result = sma03_WriteSinglePage( us, address, sg[i],address, info->rdt);
            if (result != USB_STOR_TRANSPORT_GOOD)  {
                return result;
            }
 */
    }
    return result;
}

/*
 *  Write Sectors (use READ_PAGE command)
 */
static int
sma03_WriteSectors( struct us_data *us,
                unsigned char  *buf,
                unsigned short Sectors ) {

    int  result;
    int  loop;
    unsigned int  sector;
    PBLOCKREAD_MEMORYMAP512  mapptr512;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    mapptr512 = (PBLOCKREAD_MEMORYMAP512)buf;
    for (sector = 0; sector < Sectors; sector++) {
        for (loop = 0; loop < 5; loop++ ) {
            if ( ( result = (info->write_sector)( us,
                         info->Card, mapptr512->DataArea,
                         (unsigned char *)(&mapptr512->RedundantArea) ) )
                         ) {
                sma03_ResetMedia(us);   /* if error then reset Media */
                return result;
            }
        }
        info->Card.Adrs += info->PagePerSector;
        mapptr512 ++;
    }

    return USB_STOR_TRANSPORT_GOOD;
}  /* sma03_WriteSectors */


#ifdef OLDVER
/*
 *
 */
int sma03_read_control(struct us_data *us,
		unsigned long address,
		unsigned short blocks,
		unsigned char *content,
		int use_sg) {

    // Unlike in the documentation, the last two bytes are the
    // number of blocks, not sectors.

    int result;
    unsigned short pages;

    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

//    pages = (address >> info->pageshift) & info->blockmask;
    pages = info->blocksize * blocks;
  printk("sma03_read_control: address %08lX blocks %04X  pages: %04X\n",
         address, blocks, pages);
    US_DEBUGP("Read control address %08lX blocks %04X\n",
		address, blocks);

    result = sma03_ReadPages( us, address, pages, content, use_sg);

    US_DEBUGP("Result for send_control in read_control %d\n",
		result);
  printk("sma03_read_control: sma03_ReadPages: result: %d\n", result);

    if (result != USB_STOR_TRANSPORT_GOOD)
        return result;

    return result;
}  /* sma03_read_control */
#endif

int sma03_read_deviceID(struct us_data *us,
                unsigned char *manufacturerID,
                unsigned char *deviceID) {

    int result;
    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    result = sma03_send_control(us, usb_rcvctrlpipe(us->pusb_dev, 0),
                SMA03_READ_ID,
                USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, 0,
                info->dvc, 4);
    US_DEBUGP("Result of sma03_send_control for device ID is %d\n", result);
    if (result != USB_STOR_TRANSPORT_GOOD)
        return result;

    *manufacturerID = info->dvc[0];
    *deviceID = info->dvc[1];

    return USB_STOR_TRANSPORT_GOOD;
} /* sma03_read_deviceID */

/*
 *  Check SmartMedia Write-protect
 */
static int
sma03_MediaProtect(struct us_data *us) {

    int     result;
    unsigned char buf[1];

    buf[0] = 0;

    if ( (result = sma03_send_control(us, usb_rcvctrlpipe(us->pusb_dev, 0),
                SMA03_IN_DATA,
                USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, SMA03PORT_USB_STAT,
                buf, 1)) != USB_STOR_TRANSPORT_GOOD)  {
            return result;
    }
    return ( buf[0] & SMA03_STAT_WPD );
}  /* sma03_MediaProtect */

/*
 *  Read Redundant of one block
 */
static int
sma03_ReadBlockRedu( struct us_data *us,
                union CD_ADDR address,
                unsigned char *redu_buf)
{

    int     result;
    int     loop;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    for (loop = 0; loop < 5; loop++) {
        if ( (result = sma03_send_bulkcontrol(us,
                usb_sndbulkpipe(us->pusb_dev, 3), /* Control */
                SMA03_REDU_BLOCK,
                USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                address, 0)
              ) != USB_STOR_TRANSPORT_GOOD)  {
            US_DEBUGP("sma03_ReadBlockRedu: "
                "sma03_send_bulkcontrol: %d retry: %d\n",
                sma03_ResultSave, loop +1 );
            continue;
        }

        if ( (result = sma03_bulk_transport(us,
                SMA03_DATA_READ, redu_buf,
                (info->blocksize * info->redusize),
                0) ) == USB_STOR_TRANSPORT_GOOD)  {
            break;
        }
        US_DEBUGP("sma03_ReadBlockRedu: "
               "sma03_bulk_transport: %d retry: %d\n",
               sma03_ResultSave, loop +1 );
        sma03_ResetMedia(us);   /* if error then reset Media */
    }

    return result;

}  /* sma03_ReadBlockRedu */

/*
 *  Write Redundant
 */
static int
sma03_WriteRedu( struct us_data *us,
                union CD_ADDR address,
                unsigned char *redu)
{

    int     result;
    unsigned short value;
    unsigned short index;
    unsigned char buf[2];
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    value = (unsigned short)(address.b.adr2 << 8 | address.b.adr1);
    index = (unsigned short)(address.b.adr0 << 8 | 0x00);

    if ( (result = sma03_send_control(us, usb_sndctrlpipe(us->pusb_dev, 0),
                SMA03_REDU_BLOCK,
                USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                value, index,
                NULL, 0) ) != USB_STOR_TRANSPORT_GOOD)  {
        return result;
    }

    if ( (result = sma03_bulk_transport(us, SMA03_DATA_WRITE, redu,
                info->Redud, 0)) ) {
        return result;
    }

    if ( (result = sma03_send_control(us, usb_rcvctrlpipe(us->pusb_dev, 0),
                SMA03_READ_STATUS,
                USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, 0,
                buf, 2)) != USB_STOR_TRANSPORT_GOOD)  {
        return result;
    }

    if ( buf[1] )
        return  USB_STOR_TRANSPORT_ERROR;

    return USB_STOR_TRANSPORT_GOOD;
}

/*
 *
 */
static int
sma03_WriteBlock( struct us_data *us,
                union CD_ADDR address,
                PBLOCKREAD_MEMORYMAP512  buf,
                unsigned short count)
{

    int     result;
    int     loop;
//    struct sma03_card_info *info = (struct sma03_card_info *)us->extra;

    US_DEBUGP("sma03_WriteBlock: addres: %08lX count: %04X\n", address.Adrs, count);

    for (loop = 0; loop < 5; loop++) {
        if ( (result = sma03_send_bulkcontrol(us,
                usb_sndbulkpipe(us->pusb_dev, 3), /* Control */
                SMA03_WRITE_BLOCK,
                USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                address, count)
             ) != USB_STOR_TRANSPORT_GOOD)  {
            US_DEBUGP("sma03_WriteBlock: "
                "sma03_send_bulkcontrol: %d retry: %d\n",
                sma03_ResultSave, loop +1 );
            continue;
        }

        if ( (result = sma03_bulk_transport(us,
                SMA03_DATA_WRITE,
                (unsigned char *)buf,
                sizeof(BLOCKREAD_MEMORYMAP512) * count,
                0) ) == USB_STOR_TRANSPORT_GOOD)  {
            break;
        }
        US_DEBUGP("sma03_WriteBlock: "
               "sma03_bulk_transport: %d retry: %d\n",
               sma03_ResultSave, loop +1 );
        sma03_ResetMedia(us);   /* if error then reset Media */
    }

    return result;
}  /* sma03_WriteBlock */

/*
 *
 *  return  SMA_ERROR / SMA_GOOD
 */

static int
CheckMediaStatus( struct us_data *us,
                        unsigned short *pMediaStatus  )
{
    int    result = SMA_GOOD;
    unsigned char  buf[1];

    buf[0] = 0;
    /* Read status after WriteBlockCommand(B4) */
    if ( (result = sma03_send_control(us, usb_rcvctrlpipe(us->pusb_dev, 0),
                SMA03_READ_STATUS,
                USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, 0,
                buf, 1)) != USB_STOR_TRANSPORT_GOOD)  {
        return SMA_ERROR;
    }

    *pMediaStatus = buf[0];

    if ( ( *pMediaStatus  & SM_STATUS_FAIL )
            || !(*pMediaStatus & SM_STATUS_READY) )  {
        US_DEBUGP("\n\n DANGER!!!*** \n\n"
                "*** MediaStatus Fail!! %02x\n\n",
                *pMediaStatus );
        return SMA_ERROR;
    }

    return SMA_GOOD;
}  /* CheckMediaStatus */


static int
CheckMediaStatusForWriteBlock( struct us_data *us,
                        unsigned short *pMediaStatus  )
{
    int    result = SMA_GOOD;

    /* Read status after WriteBlockCommand(B4) */
    if ( (result = sma03_send_control(us, usb_rcvctrlpipe(us->pusb_dev, 0),
                SMA03_READ_STATUS2,
                USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, 0,
                (unsigned char *)pMediaStatus,
                 2)) != USB_STOR_TRANSPORT_GOOD)  {
        return SMA_ERROR;
    }

    if ( ( *pMediaStatus  & SM_STATUS_FAIL )
            || !(*pMediaStatus & SM_STATUS_READY) )  {
        US_DEBUGP("\n\n DANGER!!!*** \n\n"
                "*** MediaStatus(for WriteBlock) Fail!! %04x\n\n",
                *pMediaStatus );
        return SMA_ERROR;
    }

    return SMA_GOOD;

}


/* ====================================================================================
 *
 *      Block Sector Write (Max 32Sector)
 *
 *    512Byte/Page ABLOCKREAD_MEMORYMAP512
 *    256Byte/Page ABLOCKREAD_MEMORYMAP256
 */
static int
new_BlockWrite( struct us_data *us,
        unsigned short  PhysicalBlock,
        unsigned short  sectors,
        short  *pMediaStatus )
{
    int      bstatus, bret;
    unsigned char *ptr_dest;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    ptr_dest = info->MyMediaBuf;

//    printk("new_BlockWrite..." );

    info->Card.Adrs = (unsigned long)PhysicalBlock << info->BitShift;

    if ( info->pagesize == Page512M ) { // 512byte/page

        bret = sma03_WriteBlock(us, info->Card,
                         (PBLOCKREAD_MEMORYMAP512)ptr_dest, sectors);
        if ( bret ) {
            US_DEBUGP("\n\n DANGER!!!*** \n\n"
                "*** BlockWriter  USB TRANSFER Fail!! %02x\n\n",
                bret );
            return bret;
        }

        // Get BlockWrite Status
        bstatus = CheckMediaStatusForWriteBlock( us, pMediaStatus );
//    printk("CheckMediaStatusForWriteBlock: %d MediaStatus: %x\n",
//                bstatus, *pMediaStatus);
        if ( ( *pMediaStatus & 0xff00 ) ) {
            bret = SMA_ERROR_TRANSFER_WRITE_BOMB;
            US_DEBUGP("\n\n DANGER!!!*** \n\n"
                "*** new_BlockWrite : USB OUT BOMBER?? \n  : "
                "UsbMediaStatus = 0x%04x \n\n",
                *pMediaStatus );
                return bret;
        }

        if ( ( *pMediaStatus & SM_STATUS_FAIL )
                 || !( *pMediaStatus & SM_STATUS_READY ) ) {
            US_DEBUGP("\n\n DANGER!!!*** \n\n"
                   "*** BlockWrite : WriteFail!!  "
                   "new_MediaStatus = %04x\n\n",
                   *pMediaStatus);
                return USB_STOR_TRANSPORT_ERROR;
        }

//        printk(" BlockWrite...return \n");
        return USB_STOR_TRANSPORT_GOOD;
    }

    /*
     * 256byte/page
     */
    if ( ( bret = sma03_WriteSectors(us, ptr_dest, sectors))
            != USB_STOR_TRANSPORT_GOOD ) {
        US_DEBUGP("\n\n DANGER!!!*** \n\n"
               "*** new_BlockWrite  USB TRANSFER Fail!! %02x\n\n",
               bret );
        return bret;
    }
    // Get BlockWrite Status
    bstatus = CheckMediaStatusForWriteBlock( us, pMediaStatus );
    if ( ( *pMediaStatus & 0xff00 ) ) {
        bret = SMA_ERROR_TRANSFER_WRITE_BOMB;
        US_DEBUGP("\n\n DANGER!!!*** \n\n"
               "*** new_BlockWrite : USB OUT BOMBER?? \n  : "
               "UsbMediaStatus = 0x%04x \n\n",
               *pMediaStatus );
        return bret;
    }

    if ( ( *pMediaStatus & SM_STATUS_FAIL )
            || !( *pMediaStatus & SM_STATUS_READY ) ) {
        US_DEBUGP("\n\n DANGER!!!*** \n\n"
               "*** new_BlockWrite : WriteFail!!  MediaStatus = %04x\n\n",
                *pMediaStatus);
        return SMA_ERROR;
    }

//    printk(" BlockWrite...return \n");

    return USB_STOR_TRANSPORT_GOOD;
}  /* new_BlockWrite */

/* ====================================================================================
 *
 *      Block Sector Write (Max 32Sector)
 *
 *    512Byte/Page ABLOCKREAD_MEMORYMAP512
 *    256Byte/Page ABLOCKREAD_MEMORYMAP256
 */

static int
old_BlockWrite( struct us_data *us,
        unsigned short  PhysicalBlock,
        unsigned short  Sectors,
        short  *pMediaStatus )

{
    int   bret = 0;
    unsigned short  sector;
    unsigned char *ptr_dest;
    PBLOCKREAD_MEMORYMAP512 mapptr512;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    US_DEBUGP("old_BlockWrite..." );

    ptr_dest = info->MyMediaBuf;
    mapptr512 = (PBLOCKREAD_MEMORYMAP512)info->MyMediaBuf;
//    info->Card.Adrs = (unsigned long)PhysicalBlock << info->BitShift;

    for ( sector = 0; sector < Sectors; sector++ ) {

        info->Card.Adrs = (unsigned long)PhysicalBlock << info->BitShift;
        info->Card.b.adr0 = info->Card.b.adr0
                         + ((unsigned short)sector * info->PagePerSector);

        bret = sma03_WriteSectors( us, (unsigned char *)mapptr512, 1);

        if ( bret != SMA_GOOD ) {
            bret = CheckMediaStatus( us, pMediaStatus );
            if ( bret == SMA_WRITE_PROTECTED ) {
                US_DEBUGP("\n DANGER!!! *** WriteProtected *** \n" );
                sma03_ResetMedia(us);   /* if error then reset Media */
                break;
            }
            if ( ( *pMediaStatus & SM_STATUS_FAIL )
                        || !( *pMediaStatus & SM_STATUS_READY ) ) {
                printk("\n\n DANGER!!!*** \n\n"
                        "*** old_BlockWrite : WriteFail!"
                        " MediaStatus=0x%04x\n\n", *pMediaStatus );
                sma03_ResetMedia(us);   /* if error then reset Media */
                break;
            }

            if ( bret == SMA_ERROR_TRANSFER_WRITE_BOMB ) {
                US_DEBUGP("\n\n DANGER!!!*** \n\n"
                        "*** old_BlockWrite : USB OUT BOMBER!!  \n\n" );
                sma03_ResetMedia(us);   /* if error then reset Media */
                break;
            }
            if ( ( *pMediaStatus & 0xff00 ) ) {
                bret = SMA_ERROR_TRANSFER_WRITE_BOMB;
                US_DEBUGP("\n\n DANGER!!!*** \n\n"
                        "*** old_BlockWrite : USB OUT BOMBER?? \n"
                        "  : UsbMediaStatus = 0x%04x \n\n",
                        *pMediaStatus );
                sma03_ResetMedia(us);   /* if error then reset Media */
                break;
            }
        }
        mapptr512 ++;
    }

    US_DEBUGP(" old_BlockWrite...return  0x%02x\n", bret );

    return bret;

}  /* old_BlockWrite */

/*
 *
 */
unsigned long
sma03_get_capacity(struct us_data *us)
{

    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);
    unsigned char manufacturerID;
    unsigned char deviceID;
    int result;

    US_DEBUGP("Reading capacity...\n");

    result = sma03_read_deviceID(us, &manufacturerID, &deviceID);

    if (result != USB_STOR_TRANSPORT_GOOD) {
        US_DEBUGP("sma03_read_deviceID is %d\n", result);
        return 0;
    }

//  printk("sma03_get_capacity: deviceID: %02X  Manuf-ID %02X\n", deviceID, manufacturerID);
    US_DEBUGP("Device ID = %02X\n", deviceID);
    US_DEBUGP("Manuf  ID = %02X\n", manufacturerID);

    if (deviceID == 0xff && manufacturerID == 0xff) {
        return 0;
    }

    info->pagesize = 512;
    info->blocksize = 16;
    info->redusize = 16;        /* Redu in page */
    info->SectorPerBlock = 16;
    info->TotalZones = 1;            /* Zone in card */
    info->PagePerSector = 1;          /* 1page(512type) or 2page(256type) */
    info->goodblocks = 1002;    /* Good blocks in zone */
    info->BitShift = 4;
    info->MaskROM = 0;
    info->ReadOnly = 0;         /* Writable */

    switch (deviceID) {

        case 0x6e: // 1MB
        case 0xe8:
        case 0xec:
            info->pagesize = 256;
            info->redusize = 8;
            info->SectorPerBlock = 16;
            info->goodblocks = 252;
            info->TotalLogicalBlock = 250;
            info->TotalPhysicalBlock = 256;
            info->PagePerSector = 2;
            return 0x00100000;

        case 0xea: // 2MB
        case 0x64:
            info->pagesize = 256;
            info->redusize = 8;
            info->SectorPerBlock = 8;
            info->goodblocks = 502;
            info->TotalLogicalBlock = 500;
            info->TotalPhysicalBlock = 512;
            info->PagePerSector = 2;
            return 0x00200000;

        case 0x5d: // 5d is a 2MB ROM card with pagesize 512.
            info->SectorPerBlock = 16;
            info->TotalLogicalBlock = 250;
            info->TotalPhysicalBlock = 256;
            info->MaskROM = 1;
            info->ReadOnly = 1;         /* Read only */
            return 0x00200000;

         // 4MB
        case 0xd5:
            info->MaskROM = 1;
            info->ReadOnly = 1;         /* Read only */
        case 0xe3:
        case 0xe5:
        case 0x6b:
            info->SectorPerBlock = 16;
            info->goodblocks = 502;
            info->TotalLogicalBlock = 500;
            info->TotalPhysicalBlock = 512;
            return 0x00400000;

         // 8MB
        case 0xd6: // ROM
            info->MaskROM = 1;
            info->ReadOnly = 1;         /* Read only */
        case 0xe6:
            info->SectorPerBlock = 16;
            info->TotalLogicalBlock = 1000;
            info->TotalPhysicalBlock = 1024;
            return 0x00800000;

         // 16MB
        case 0x57:  // ROM
            info->MaskROM = 1;
            info->ReadOnly = 1;         /* Read only */
        case 0x73:
            info->blocksize = 32;
            info->SectorPerBlock = 32;
            info->TotalLogicalBlock = 1000;
            info->TotalPhysicalBlock = 1024;
            info->BitShift = 5;
            return 0x01000000;

         // 32MB
        case 0x58: // ROM
            info->MaskROM = 1;
            info->ReadOnly = 1;         /* Read only */
        case 0x75:
            info->blocksize = 32;
            info->SectorPerBlock = 32;
            info->TotalZones = 2;
            info->TotalLogicalBlock = 2000;
            info->TotalPhysicalBlock = 2048;
            info->BitShift = 5;
            return 0x02000000;

         // 64MB
        case 0xD9: // ROM
            info->MaskROM = 1;
            info->ReadOnly = 1;         /* Read only */
        case 0x76:
            info->blocksize = 32;
            info->SectorPerBlock = 32;
            info->TotalZones = 4;
            info->TotalLogicalBlock = 4000;
            info->TotalPhysicalBlock = 4096;
            info->BitShift = 5;
            return 0x04000000;

         // 1284MB
        case 0xDA: // ROM
            info->MaskROM = 1;
            info->ReadOnly = 1;         /* Read only */
        case 0x79:
            info->blocksize = 32;
            info->SectorPerBlock = 32;
            info->TotalZones = 8;
            info->TotalLogicalBlock = 8000;
            info->TotalPhysicalBlock = 8192;
            info->BitShift = 5;
            return 0x08000000;

        default: // unknown
            return 0;

    }
}  /* sma03_get_capacity */

/***************************************************
    Generate C Parity ECC
    Record to ecctbl[]
****************************************************/
static void GenCPtbl(struct us_data *us)
{
    unsigned int ii,jj;
    unsigned char dall;                        /* D-all(XOR all bit) */
    unsigned char cp0,cp1,cp2,cp3,cp4,cp5;      /* CP0,CP1,CP2,... */
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    for(ii=0; ii<256; ii++)
        info->ecctbl[ii] = 0x00;  /* 0 Clear */
    for(ii=0; ii<256; ii++) {
        dall = cp0 = cp1 = cp2 = cp3 = cp4 = cp5 = 0x00;
        for(jj=0; jj<7; ++jj) {
            if((unsigned char)ii & BIT0){++dall;++cp0;++cp2;++cp4;}
            if((unsigned char)ii & BIT1){++dall;++cp1;++cp2;++cp4;}
            if((unsigned char)ii & BIT2){++dall;++cp0;++cp3;++cp4;}
            if((unsigned char)ii & BIT3){++dall;++cp1;++cp3;++cp4;}
            if((unsigned char)ii & BIT4){++dall;++cp0;++cp2;++cp5;}
            if((unsigned char)ii & BIT5){++dall;++cp1;++cp2;++cp5;}
            if((unsigned char)ii & BIT6){++dall;++cp0;++cp3;++cp5;}
            if((unsigned char)ii & BIT7){++dall;++cp1;++cp3;++cp5;}
        }
        if(dall & BIT0) info->ecctbl[ii]|=BIT6;
        if(cp5 & BIT0) info->ecctbl[ii]|=BIT5;
        if(cp4 & BIT0) info->ecctbl[ii]|=BIT4;
        if(cp3 & BIT0) info->ecctbl[ii]|=BIT3;
        if(cp2 & BIT0) info->ecctbl[ii]|=BIT2;
        if(cp1 & BIT0) info->ecctbl[ii]|=BIT1;
        if(cp0 & BIT0) info->ecctbl[ii]|=BIT0;
    }
}

unsigned short
LogicalBlockToBlockAddressField( struct us_data *us,
                        unsigned short LogicalBlock )
{
    unsigned int ll;
    unsigned int msk,prt;
    unsigned short      BlockAddressFieldData;

    msk = 0x0002;
    BlockAddressFieldData = (LogicalBlock << 1) | 0x1000;

    for ( prt=0,ll=0; ll < 15; ll++ ) {
        if ( BlockAddressFieldData & (msk << ll) )
            ++prt;
    }
    BlockAddressFieldData |= (prt & 0x0001);

    BlockAddressFieldData = EX_HILO_WORD( BlockAddressFieldData );

    return BlockAddressFieldData;
}

/************************************************
 Read Hidden Block
************************************************/
int HiddRd(struct us_data *us)
{
    int result;
    union CD_ADDR   hid_addr;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    if (info->dvc[0] == 0xEC) { /* SAMSUNG */
        if ((result = sma03_send_control(us, usb_sndctrlpipe(us->pusb_dev, 0),
                        SMA03_OUT_DATA,
                        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                        SMA03_MODE_WR_CMD, SMA03PORT_USB_MODE,
                        NULL, 0)) != USB_STOR_TRANSPORT_GOOD)  {
            return result;
        }
        if ((result = sma03_send_control(us, usb_sndctrlpipe(us->pusb_dev, 0),
                        SMA03_OUT_DATA,
                        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                        0x30, SMA03PORT_USB_DATA,
                        NULL, 0)) != USB_STOR_TRANSPORT_GOOD)  {
            return result;
        }
        if ((result = sma03_send_control(us, usb_sndctrlpipe(us->pusb_dev, 0),
                        SMA03_OUT_DATA,
                        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                        0x65, SMA03PORT_USB_DATA,
                        NULL, 0)) != USB_STOR_TRANSPORT_GOOD)  {
            return result;
        }
        if ((result = sma03_send_control(us, usb_sndctrlpipe(us->pusb_dev, 0),
                        SMA03_OUT_DATA,
                        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                        SMA03_MODE_WR_DATA, SMA03PORT_USB_MODE,
                        NULL, 0)) != USB_STOR_TRANSPORT_GOOD)  {
            return result;
        }
        hid_addr.Adrs = 0L;
        result = sma03_ReadSinglePage( us, hid_addr,
                                        info->hid_buf, info->hid_rdt);
   } else {
        if ((result = sma03_send_control(us, usb_sndctrlpipe(us->pusb_dev, 0),
                        SMA03_READ_HIDDEN,
                        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                        0, 0,
                        NULL, 0)) != USB_STOR_TRANSPORT_GOOD)  {
            return result;
        }
        /* Read data */
        if ((result = sma03_bulk_transport(us, SMA03_DATA_READ,
                        info->hid_buf, info->pagesize,
                        0)) != USB_STOR_TRANSPORT_GOOD)  {
            return result;
        }
        /* Read redu */
        result = sma03_bulk_transport(us, SMA03_DATA_READ,
                        info->hid_rdt, info->redusize,
                        0);
    }
//  printk("HiddneData:\n");
//  { int i; for (i = 0; i<16; i++) printk("%02X ", info->hid_buf[i]); } /* */
//  printk("HiddneRedu:\n");
//  { int i; for (i = 0; i<16; i++) printk("%02X ", info->hid_rdt[i]); } /* */
    if (result != USB_STOR_TRANSPORT_GOOD)
            return result;

    sma03_ResetMedia(us);

    return USB_STOR_TRANSPORT_GOOD;

}
/***********************************************
        Check for Prity
***********************************************/
static int
CheckBlockAddressFieldParity( unsigned char *BlockAddressFieldData )
{
    int parity_flag;
    unsigned short  ll;
    unsigned short msk;
    int parity_count;
    unsigned short blockfield;

    blockfield = short_pack2( BlockAddressFieldData );

    parity_flag = blockfield & 0x0001;

    msk = 0x0001;

    parity_count = 0;
    for ( ll=1; ll < 16; ll++ ) {
        if ( blockfield & (msk << ll) ) {
            parity_count ++;
        }
    }
    parity_count &= 0x01;

    if ( parity_flag == parity_count )
        return SMA_GOOD;
    else
        return SMA_ERROR;

}

/***********************************************
        Check for 0bit Count
***********************************************/
static int Bit0Ck(unsigned char dat,unsigned char ckc)
{
    unsigned char msk, bct;

    bct = 0x00;
    msk = 0x01;
    do{
        if((dat & msk)==0){
            if(++bct >= ckc) break;
        }
    }while((msk<<=1));

    return(msk);
}  /* Bit0Ck */

/***************************************************
    Calculating ECC
    using CP0-CP5 code ecctbl[]
****************************************************/
static void
CalculateEcc(struct us_data *us, unsigned char *buf)
{
    unsigned int i;
//    int  c0, c1;

    unsigned int cnt;	/* Loop Counter  */
    unsigned char wec1;	/* LP14,LP12,LP10,... */
    unsigned char wec2;	/* LP14,LP12,LP10,... */
    unsigned char wec3;	/* LP15,LP13,LP11,... */
    unsigned char msk0,msk1;
    unsigned char ecc1;	/* LP15,LP14,LP13,... */
    unsigned char ecc2;	/* LP07,LP06,LP05,... */

    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    for (cnt=0; cnt < 2 ;cnt++) {
        wec1 = wec2 = wec3 = 0;
        for (i=0; i < 256 ;i++) {
/*
            c0 = *buf & 0x00ff;
            c1 = *buf++ & 0x00ff;
            wec1 ^= (info->ecctbl[c0] & 0x3f);
            if ((info->ecctbl[c1] & BIT6)) {
                wec3 ^= (unsigned char)i;
                wec2 ^= ~((unsigned char)i);
            }
*/
            wec1 ^= (info->ecctbl[*buf] & 0x3f);
            if ((info->ecctbl[(*buf++)] & BIT6)) {
                wec3 ^= (unsigned char)i;
                wec2 ^= ~((unsigned char)i);
            }

        }  /* for (i = 0...256 ) */
        msk0 = BIT7; msk1 = BIT7;
        ecc1 = ecc2 = 0x00;
        for (i=0; i < 4; ++i)  {
            if ((wec3 & msk0) != 0)  ecc1|=msk1;
            msk1 >>= 1;
            if ((wec2 & msk0)!=0)    ecc1|=msk1;
            msk1 >>= 1; msk0 >>= 1;
        }  /* for (i = 0...4 ) */
        msk1=BIT7;
        for (i=0;i < 4;++i) {
            if ((wec3 & msk0)!=0)    ecc2|=msk1;
            msk1 >>= 1;
            if ((wec2 & msk0)!=0)    ecc2|=msk1;
            msk1 >>= 1; msk0 >>= 1;
        }  /* for (i = 0...4 ) */

        i = cnt * 3;	/* cnt = 0 or 1 */
        info->Ecc[(0+i)] = (unsigned char)~ecc2;
        info->Ecc[(1+i)] = (unsigned char)~ecc1;
        info->Ecc[(2+i)] = (unsigned char)((~wec1 << 2) | BIT0011);
    }  /* for (cnt = 0...2 ) */
}  /* CalculateEcc */

// ***************************************************
//   Correct Data (Read Only?j
//  using CP0-CP5 code ecctbl[]
// ****************************************************/

static int
CorrectData( struct us_data *us,
        unsigned char *buf,
        unsigned int cnt )
{
    register unsigned int i,ii; /* For counting */
//    int  status;
    unsigned long wpy;	/* Working to check pya  */
    unsigned long pya;	/* Result of comparison  */
    unsigned char py1,py2,py3; /* Result of comparison */
    unsigned char wby;	/* Working for erbyt */
    unsigned char erbyt;	/* Byte address of cor. DATA */
    unsigned char wbi;	/* Working for erbit */
    unsigned char erbit;	/* Bit address of cor. DATA  */

    unsigned char *redundant;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    redundant = (unsigned char *) &(((PBLOCKREAD_MEMORYMAP512)buf)->RedundantArea);

    printk("CorrectData\n");

    i = cnt * 5;
    ii = cnt * 3;
    py1 = info->Ecc[(1+ii)] ^ redundant[(ECA11-i)];
    py2 = info->Ecc[(0+ii)] ^ redundant[(ECA10-i)];
    py3 = info->Ecc[(2+ii)] ^ redundant[(ECA12-i)];
    pya = ((unsigned long)py1<<16)
            +((unsigned long)py2<<8)
            +(unsigned long)py3;
    if ( ((pya^(pya >> 1)) & CORRECTABLE) == CORRECTABLE ) {
        erbyt = 0;
        wpy=BIT23;
        wby=BIT7;
        for (i=0;i < 8;++i) { // Checking 8 bit
            if ( (pya & wpy)!=0 )
                erbyt|=wby;	/* from LP's*/
            wpy>>=2;
            wby>>=1;
        } /* for (i=0...8) */
        erbit = 0;
        wbi=BIT2;
        for ( i=0;i < 3;++i ) { /* Checking 3 bit */
            if ((pya & wpy)!=0)
                erbit|=wbi;     /* from CP's*/
            wpy>>=2;
            wbi>>=1;
        }
        wbi = BIT0;
        i = (cnt * 256) + (unsigned int)erbyt;
        buf[i] ^= (wbi << erbit);       // Put corrected data
        return SMA_GOOD;
    }

    i = cnt * 5;	// Set ECC
    redundant[(ECA11-i)] = info->Ecc[(1+ii)];
    redundant[(ECA10-i)] = info->Ecc[(0+ii)];
    redundant[(ECA12-i)] = info->Ecc[(2+ii)];

    i = 0;
    pya &= 0x00ffffffL; /* Masking*/
    while (pya) {  /* If pya=0 finish counting*/
        if (pya & BIT0)
            ++i;        /* Count number of 1 bit*/
        pya>>=1;
    }   /* ECC Error */

    if (i==1)
        return SMA_GOOD;

    return SMA_ERROR;  // correctable error (NG)

}

/***************************************************
	Ecc Check Only
****************************************************/
static int
EccCodeCheck(struct us_data *us, unsigned char * buf, unsigned int RW_Flag )
{
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    CalculateEcc( us, buf ); /* Calc. ECC */


    if ( RW_Flag ) {  /* Write */
        info->rdt[ECA10] = info->Ecc[0];
        info->rdt[ECA11] = info->Ecc[1];
        info->rdt[ECA12] = info->Ecc[2];
        info->rdt[ECA20] = info->Ecc[3];
        info->rdt[ECA21] = info->Ecc[4];
        info->rdt[ECA22] = info->Ecc[5];
    } else {  /* Read  */
        if ( ((info->Ecc[2]^info->rdt[ECA12])
                + (info->Ecc[1]^info->rdt[ECA11])
                + (info->Ecc[0]^info->rdt[ECA10])) ) {
            printk("EccCodeCheck 000-255\n");
#if DBG
{
int	i;
for(i = 0; i < SECTOR; i++)
    if(info->buf[i] != (i&0xff))
        printk("buf[%x] = 0x%x\n", i, info->buf[i]);
        printk("Ecc[0] = 0x%x\n", info->Ecc[0]);
        printk("Ecc[1] = 0x%x\n", info->Ecc[1]);
        printk("Ecc[2] = 0x%x\n", info->Ecc[2]);
        printk("rdt[ECA10] = 0x%x\n", info->rdt[ECA10]);
        printk("rdt[ECA11] = 0x%x\n", info->rdt[ECA11]);
        printk("rdt[ECA12] = 0x%x\n", info->rdt[ECA12]);
        printk("buf = %p\n", info->buf));
}
#endif
            return SMA_ERROR;
        }

        if ( ((info->Ecc[5]^info->rdt[ECA22])
                + (info->Ecc[4]^info->rdt[ECA21])
                + (info->Ecc[3]^info->rdt[ECA20]))){
            printk("EccCodeCheck 256-512\n");
#if	DBG
{
int	i;
for(i = 0; i < SECTOR; i++)
    if(info->buf[i] != (i&0xff))
        printk("buf[%x] = 0x%x\n", i, info->buf[i]);
        printk("Ecc[3] = 0x%x\n", info->Ecc[3]);
        printk("Ecc[4] = 0x%x\n", info->Ecc[4]);
        printk("Ecc[5] = 0x%x\n", info->Ecc[5]);
        printk("rdt[ECA20] = 0x%x\n", info->rdt[ECA20]);
        printk("rdt[ECA21] = 0x%x\n", info->rdt[ECA21]);
        printk("rdt[ECA22] = 0x%x\n", info->rdt[ECA22]);
        printk("buf = %p\n", info->buf));
}
#endif
            return SMA_ERROR;
        }
    }
    return SMA_GOOD;
}

/**************************************************
    Ecc Handling
***************************************************/
static int
EccCodeHandSingle(struct us_data *us,
                unsigned char *buf,
                unsigned char *redu,
                unsigned int RW_Flag )
{
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);
    PREDUNDANT_AREA redundant_ptr;
    redundant_ptr = (PREDUNDANT_AREA)redu;

    CalculateEcc( us, buf ); /* Calc. ECC */

    if ( RW_Flag ) { /* Write */
        redundant_ptr->EccArea1[0] = info->Ecc[0];
        redundant_ptr->EccArea1[1] = info->Ecc[1];
        redundant_ptr->EccArea1[2] = info->Ecc[2];
        redundant_ptr->EccArea2[0] = info->Ecc[3];
        redundant_ptr->EccArea2[1] = info->Ecc[4];
        redundant_ptr->EccArea2[2] = info->Ecc[5];
    } else { /* Read */
        if ( (info->Ecc[2] != redundant_ptr->EccArea1[2])
           || (info->Ecc[1] != redundant_ptr->EccArea1[1])
           || (info->Ecc[0] != redundant_ptr->EccArea1[0]) ) {
            printk ("EccCodeHand 000-255\n");
            return SMA_ERROR;
        }

        if ( (info->Ecc[5] != redundant_ptr->EccArea2[2])
           || (info->Ecc[4] != redundant_ptr->EccArea2[1])
           || (info->Ecc[3] != redundant_ptr->EccArea2[0]) ) {
            printk("EccCodeHand 256-512\n");
            return SMA_ERROR;
        }
    }
    return SMA_GOOD;
}


static int
EccCodeHand(struct us_data *us, unsigned char *buf, unsigned int RW_Flag )
{
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);
    PREDUNDANT_AREA redundant_ptr;
    redundant_ptr = &(((PBLOCKREAD_MEMORYMAP512)buf)->RedundantArea);

    CalculateEcc( us, buf ); /* Calc. ECC */

    if ( RW_Flag ) { /* Write */
        redundant_ptr->EccArea1[0] = info->Ecc[0];
        redundant_ptr->EccArea1[1] = info->Ecc[1];
        redundant_ptr->EccArea1[2] = info->Ecc[2];
        redundant_ptr->EccArea2[0] = info->Ecc[3];
        redundant_ptr->EccArea2[1] = info->Ecc[4];
        redundant_ptr->EccArea2[2] = info->Ecc[5];
    } else { /* Read */
        if ( (info->Ecc[2] != redundant_ptr->EccArea1[2])
           || (info->Ecc[1] != redundant_ptr->EccArea1[1])
           || (info->Ecc[0] != redundant_ptr->EccArea1[0]) ) {
            printk ("EccCodeHand 000-255\n");
            printk("EccC: %02X %02X %02X  EccR: %02X %02X %02X\n",
                        info->Ecc[2], info->Ecc[1], info->Ecc[0],
                        redundant_ptr->EccArea1[2],
                        redundant_ptr->EccArea1[1],
                        redundant_ptr->EccArea1[0] );
            return SMA_ERROR;
        }

        if ( (info->Ecc[5] != redundant_ptr->EccArea2[2])
           || (info->Ecc[4] != redundant_ptr->EccArea2[1])
           || (info->Ecc[3] != redundant_ptr->EccArea2[0]) ) {
            printk("EccCodeHand 256-512\n");
            return SMA_ERROR;
        }
    }
    return SMA_GOOD;
}

/************************************************
     CIS Infomation Check
************************************************/
int ReadCheckCis(unsigned char *com)
{
    register unsigned int cm;
    unsigned char id[]={0x01,0x03,0xd9,0x01,0xff,0x18,0x02,0xdf,0x01,0x20};

    for(cm=0;cm < 10;cm++){
        if( com[cm] != id[cm] )
            return(SMA_ERROR);
    }
    return(SMA_GOOD);
}  /* ReadCheckCis */


static int
CisSearchAndCheck( struct us_data *us,
                unsigned short *pCisBlock )
{
    REDUNDANT_AREA* redundant;
    unsigned long SaveCardAdr;
    int  iret;
    int  bret = 0;
    unsigned short block;
    unsigned short sector;
    int  loop;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    SaveCardAdr = info->Card.Adrs;
    info->Card.Adrs = 0;

    sma03_ResetMedia(us);

    for ( block = 0; block < 23; block++ ) {
        for ( sector = 0; sector < 8; sector++ ) {
            info->Card.Adrs = block << info->BitShift;
            info->Card.b.adr0 += ((int)sector * info->PagePerSector);

            for (loop = 0; loop < 5; loop ++ ) {
                bret = info->read_sector( us,
                                info->Card, info->cis_buf, info->cis_rdt);
                if ( bret == USB_STOR_TRANSPORT_GOOD)
                    break;
            }
            if ( bret != USB_STOR_TRANSPORT_GOOD ) {
                US_DEBUGP("sma03_ReadSector ReadFail  "
                        "Block=0x%x, sector=%02d \n", block , sector);
                info->Card.Adrs = SaveCardAdr;
                return SMA_ERROR;
            }

            redundant = (PREDUNDANT_AREA) info->cis_rdt;
            if ( redundant->BlockStatusArea != 0xff ) {
                if ( Bit0Ck( redundant->BlockStatusArea, 2 ) ) {
                    US_DEBUGP("  *** BlockStatus Bad Block!! : "
                           "PhysicalBlock=0x%x : BlockStatus=%02x\n",
                           block, (unsigned char)(redundant->BlockStatusArea) );
                    break;
                }
            }

            if ( redundant->DataStatusArea != 0xff ) {
                if ( Bit0Ck( redundant->DataStatusArea, 4) ) {
                    US_DEBUGP("  *** DataStatus : Bad Sector!! : "
                           "PhysicalBlock=0x%x : sector=0x%x : DataStatus=%02x\n",
                           block, sector, (unsigned char)(redundant->DataStatusArea) );
                    continue;
                }
            }

            CalculateEcc( us, info->cis_buf ); /* Calc. ECC */

            if ( (info->Ecc[2] != redundant->EccArea1[2])
                    || (info->Ecc[1] != redundant->EccArea1[1])
                    || (info->Ecc[0] != redundant->EccArea1[0]) ) {
                US_DEBUGP("CisCheck ECC 000-255 Fail  "
                        "Block=0x%x, sector=%02d \n", block, sector );
                iret = CorrectData( us, info->cis_buf, 0 /*0-255*/ );
                if ( iret == SMA_GOOD ) {
                    if ( ReadCheckCis( info->cis_buf )
                                 == SMA_GOOD ) {
                        *pCisBlock = block;
                        info->Card.Adrs = SaveCardAdr;
                        return SMA_GOOD;
                    } else {
                        US_DEBUGP(" ReadCheckCis NoGood  "
                                "Block=0x%x, sector=%02d \n",
                                block , sector);
                        info->Card.Adrs = SaveCardAdr;
                        return SMA_ERROR;
                    }
                }  /* iret == OK */
              /* ECC error */
            } else {
                if ( ReadCheckCis( info->cis_buf )
                                 == SMA_GOOD ) {
                    *pCisBlock = block;
                    info->Card.Adrs = SaveCardAdr;
                    return SMA_GOOD;
                } else {
                    US_DEBUGP(" ReadCheckCis NoGood  "
                           "Block=0x%x, sector=%02d \n",
                           block , sector );
                    info->Card.Adrs = SaveCardAdr;
                    return SMA_GOOD;
                }
            }

            if ( (info->Ecc[5] != redundant->EccArea2[2])
                    || (info->Ecc[4] != redundant->EccArea2[1])
                    || (info->Ecc[3] != redundant->EccArea2[0]) ) {
                US_DEBUGP("CisCheck ECC 256-511 Fail  "
                        "Block=0x%x, sector=%02d \n",
                        block, sector );

                iret = CorrectData( us, info->cis_buf, 1 /*256-511*/ );
                if ( ReadCheckCis( info->cis_buf + 256)
                                 == SMA_GOOD ) {
                    *pCisBlock = block;
                    info->Card.Adrs = SaveCardAdr;
                    return SMA_GOOD;
                } else {
                    US_DEBUGP(" ReadCheckCis NoGood  "
                           "Block=0x%x, sector=%02d \n",
                           block , sector );
                    info->Card.Adrs = SaveCardAdr;
                    return SMA_ERROR;
                }
              /* ECC error */
            } else {
                if ( ReadCheckCis( info->cis_buf + 256 )
                                 == SMA_GOOD ) {
                    *pCisBlock = block;
                    info->Card.Adrs = SaveCardAdr;
                    return SMA_GOOD;
                } else {
                    US_DEBUGP(" ReadCheckCis NoGood  "
                           "Block=0x%x, sector=%02d \n",
                           block , sector );
                    info->Card.Adrs = SaveCardAdr;
                    return SMA_ERROR;
                }
            }
        }  /* for (sector = 0...8) */
    }  /* for (block = 0...23) */

    US_DEBUGP(" CisSearch : can't Find CisBlock\n" );

    info->Card.Adrs = SaveCardAdr;
    return SMA_ERROR;

}  /* CisSearchAndCheck */


static int
BadBlockSet( struct us_data *us,
                unsigned short BadPhysicalBlock )
{
    PREDUNDANT_AREA redundant_ptr;
    int    Zone;
    unsigned long  SaveCardAdr;
    unsigned long  iii;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    Zone =  BadPhysicalBlock / MAX_BLOCK_PER_ZONE_PHYSICAL /* 1024 */ ;

  // Count Up Bad Block
    info->BadBlockCountPerZone[Zone] ++;

    US_DEBUGP("\n *** BadBlock Set !!! *\n"
        "   BadBlockCount[Zone:%d] = %d,  BadPhisicalBlock = 0x%x\n",
        Zone, info->BadBlockCountPerZone[Zone], BadPhysicalBlock );

    if ( info->ReadOnly ) {
        return SMA_ERROR;
    }

    SaveCardAdr = info->Card.Adrs;
    info->Card.Adrs = (unsigned long)BadPhysicalBlock << info->BitShift;

    redundant_ptr = (PREDUNDANT_AREA)info->cpy_rdt;

    memset( (unsigned char *)redundant_ptr, 0xff, info->SectorPerBlock * 16);

    for ( iii=0; iii < info->SectorPerBlock; iii++ ) {
        redundant_ptr->BlockStatusArea = 0xf0; // BlockStatusFlag = 0xf0
        if ( sma03_WriteRedu(us, info->Card, info->cpy_rdt) )
            return (SMA_ERROR);
        info->Card.Adrs += info->PagePerSector;
        redundant_ptr ++;
    }
    info->Card.Adrs = SaveCardAdr;

    sma03_ResetMedia(us);

    return SMA_GOOD;
}

static int
CheckWriteEntryTable( struct us_data *us )
{
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);
    unsigned short const  Zone = info->ac_zone;
    unsigned short  block;
    unsigned short  lll;
    unsigned short  size;
    int  ex_flag;

    US_DEBUGP("*** CheckWriteEntryTable *** Zone = %d \n",Zone);

    if ( info->BadBlockCountPerZone[Zone] > info->LimitBadBlock[Zone] ) {
        printk("?@CheckWriteEntryTable :  BadBlock > Limit !!\n");
        return SMA_ERROR;
    }

    ex_flag = FALSE;
    size = info->tblsz[Zone];

    for ( lll = 0; lll < size; lll++ ) {
        block = info->WriteEntryTableList[ Zone ][ info->WritePoint[Zone] ];
        if ( block != 0xffff ) {
            ex_flag = TRUE;
            break;
        }

        info->WritePoint[Zone]++;
        if ( info->WritePoint[Zone] >= size )
            info->WritePoint[Zone] = 0;

    }
    if ( ex_flag == FALSE )  {
        printk("CheckWriteEntryTable : No AKI Block int WriteEntryTableList!!\n");
        return SMA_ERROR;
    }

    block = info->WriteEntryTableList[ Zone ][ info->WritePoint[ Zone ] ];
    if ( block == info->BlockTable[ Zone ][ info->ac_block ] ) {
        printk("CheckWriteEntryTable : "
                "Sude Ni BlockTable ni sonzai!! : "
                "Block = 0x%x\n", block);
        return SMA_ERROR;
    }

    return SMA_GOOD;

}
/***********************************************
        Block Erase
***********************************************/
static int
sma03New_EraseBlock(struct us_data *us,
                unsigned short DeletePhysicalBlock,
                unsigned short *pMediaStatus )
{
    int    result;
    unsigned char  buf[1];
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);


    if ( info->ReadOnly ) {
        return SMA_WRITE_PROTECTED;
    }

    info->Card.Adrs = (unsigned long)DeletePhysicalBlock << info->BitShift;

    /* Erase Block of SmartMedia */
    if ( ( result = sma03_send_bulkcontrol(us,
                    usb_sndbulkpipe(us->pusb_dev, 3),
                    SMA03_ERASE_BLOCK,
                    USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                    info->Card, 0) ) ) {
        printk("sma03New_EraseBlock: sma03_send_bulkcontrol:%d\n",
                sma03_ResultSave);
        return(SMA_ERROR);
    }

    buf[0] = 0;
    if ( ( result = sma03_send_control(us,
                usb_rcvctrlpipe(us->pusb_dev, 0),
                SMA03_READ_STATUS,
                USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, 0,
                buf, 1) ) ) {
        printk("sma03New_EraseBlock: sma03_send_control:%d\n",
                sma03_ResultSave);
        return(SMA_ERROR);
    }

    *pMediaStatus = (unsigned short)(buf[0]);
    if ( buf[0] & WR_FAIL ) {
        printk("EraseBlock Busy\n");
        sma03_ResetMedia(us);   /* if error then reset Media */
        return(SMA_ERROR);
    }

    if ( ! (buf[0] & SM_STATUS_READY) ) { /* b6 */
        printk("BlockErase :  ERASE FAIL: BUSY!!! [status=%x]\n",
            *pMediaStatus);
        sma03_ResetMedia(us);   /* if error then reset Media */
        return(SMA_ERROR);
    }

    return SMA_GOOD;

}  /* sma03_NewEraseBlock */

/***********************************************
        Block Erase
***********************************************/
static int
sma03Old_EraseBlock(struct us_data *us,
                unsigned short DeletePhysicalBlock,
                unsigned short *pMediaStatus )
{
    unsigned char  buf[1];
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);


    if ( info->ReadOnly ) {
        return SMA_WRITE_PROTECTED;
    }

    if( sma03_MediaProtect(us) ) { /* Check Write-protect */
        US_DEBUGP("sma03Old_EraseBlock: sma03_MediaProtect:%d\n",
                sma03_ResultSave);
        return(SMA_WRITE_PROTECTED);
    }

    info->Card.Adrs = (unsigned long)DeletePhysicalBlock << info->BitShift;
    buf[0] = 0;

    /* Erase Block of SmartMedia */
    if ( sma03_send_bulkcontrol(us,
                    usb_sndbulkpipe(us->pusb_dev, 3),
                    SMA03_ERASE_BLOCK,
                    USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                    info->Card, 0) ) {
        US_DEBUGP("sma03Old_EraseBlock: sma03_send_bulkcontrol:%d\n",
                sma03_ResultSave);
        return(SMA_ERROR);
    }

    if ( sma03_send_control(us, usb_rcvctrlpipe(us->pusb_dev, 0),
                SMA03_READ_STATUS,
                USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                0, 0,
                buf, 1) ) {
        US_DEBUGP("sma03Old_EraseBlock: sma03_send_control:%d\n",
                sma03_ResultSave);
        return(SMA_ERROR);
    }

    *pMediaStatus = buf[0];
    if ( buf[0] & SM_STATUS_FAIL ) {
        printk("EraseBlock Busy\n");
        if ( sma03_StanbyMode(us) )  /* Set Stanby mode */
            return(SMA_DEVICE_ERROR);
        sma03_ResetMedia(us);   /* if error then reset Media */
        return(SMA_ERROR);
    }
    if ( !( buf[0] & SM_STATUS_READY /* b6 */ ) ) {
        printk("sma03Old_EraseBlock:  ERASE FAIL: BUSY!!! [status=%x]\n",
                    *pMediaStatus);
        if ( sma03_StanbyMode(us) )  /* Set Stanby mode */
            return(SMA_DEVICE_ERROR);
        sma03_ResetMedia(us);   /* if error then reset Media */
        return(SMA_ERROR);
    }

    if ( sma03_StanbyMode(us) )  /* Set Stanby mode */
        return(SMA_DEVICE_ERROR);

    return SMA_GOOD;

}  /* sma03Old_EraseBlock */


static int
EraseForWriteRetry( struct us_data *us,
                unsigned short DeleteBlock,
                unsigned short *pMediaStatus )
{
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    sma03_ResetMedia(us);   /* reset Media */

    return ( info->erase_block( us, DeleteBlock, pMediaStatus ) );
}  /* EraseForWriteRetry */

int
EraseBlock( struct us_data *us,
                unsigned short DeleteBlock)
{
    int     bstatus;
    unsigned short  MediaStatus;
    unsigned short  Zone = DeleteBlock / MAX_BLOCK_PER_ZONE_PHYSICAL /*1024*/;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    US_DEBUGP(" Erase Block ...  Delete Block = 0x%x\n", DeleteBlock );

    if ( info->WriteEntryTableList[ Zone ][ info->ErasePoint[Zone] ]
                != 0xffff ) {
        US_DEBUGP("This block is already registered as Deleted Block : \n");
        US_DEBUGP("  Zone= %d,  ErasePoint= 0x%x, WriteEntryTable[][]= 0x%x \n",
                Zone,
                info->ErasePoint[ Zone ],
                info->WriteEntryTableList[ Zone ][ info->ErasePoint[Zone] ] );
        /*
         *  set MEDIA_ERROR
         */
        return SMA_MEDIA_ERROR;
    }

    //info->Card.Adrs = (unsigned long)DeleteBlock << info->BitShift;
    bstatus = info->erase_block( us, DeleteBlock, &MediaStatus );
    if ( bstatus == SMA_WRITE_PROTECTED ) {
        /*
         *  set WRITE_PROTECTED
         */
        return bstatus;
    }

  US_DEBUGP("EraseBlock: "
        "info->erase_block: %d MediaStatus: %x\n",
        bstatus, MediaStatus);
    if ( bstatus ) {
        if ( !( MediaStatus & SM_STATUS_READY /* b6 */ ) ) {
            /*
             *  set DEVICE_ERROR
             */
            return SMA_DEVICE_ERROR;
        }
        if ( MediaStatus & SM_STATUS_FAIL ) {
            US_DEBUGP("\n EraseFail => Bad Block !!! ** : "
                "PhysicalBlock= 0x%x : BadBlockCount = %d\n",
                DeleteBlock, info->BadBlockCountPerZone[Zone]+1 );
                BadBlockSet( us, DeleteBlock );
        }

        return SMA_GOOD;
    }

    /* Register to WriteEntryTableList */
    info->WriteEntryTableList[ Zone ][ info->ErasePoint[Zone] ] = DeleteBlock;
    info->ErasePoint[ Zone ] ++ ;
    if ( info->ErasePoint[Zone] >= info->tblsz[Zone] ) {
        info->ErasePoint[ Zone ] = 0;
    }

    return SMA_GOOD;
}  /* EraseBlock */



static int
CheckGetBlockAddressSub( struct us_data *us,
                        REDUNDANT_AREA* redundant,
                        unsigned short* LogicalBlock )
{
    int result;
    int ParityCheck1;
    int ParityCheck2;
    unsigned short LogicalBlocksPerZone;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);
    unsigned short LogicalBlockAddress1;
    unsigned short LogicalBlockAddress2;

    LogicalBlockAddress1 = short_pack2(redundant->BlockAddressArea1);
    LogicalBlockAddress2 = short_pack2(redundant->BlockAddressArea2);

//    unsigned short LogicalBlockAddress1 = EX_HILO_WORD(redundant->BlockAddressArea1);
//    unsigned short LogicalBlockAddress2 = EX_HILO_WORD(redundant->BlockAddressArea2);

    ParityCheck1 = CheckBlockAddressFieldParity( redundant->BlockAddressArea1 );
    ParityCheck2 = CheckBlockAddressFieldParity( redundant->BlockAddressArea2 );
    LogicalBlocksPerZone = (unsigned short)info->TotalLogicalBlock/info->TotalZones;

    result = SMA_ERROR;

    /* redundant->BlockAddressArea1 == redundant->BlockAddressArea2 */
    if ( LogicalBlockAddress1 == LogicalBlockAddress2 ) {
        if ( ParityCheck1 == SMA_GOOD ) {
            *LogicalBlock = (LogicalBlockAddress1 & 0x07fe) >> 1;
            if ( *LogicalBlock < LogicalBlocksPerZone ) {
                result = SMA_GOOD;
            } else {
                printk("1 LogicalBlockAddress[0x%x] > 0x%x OverError!! \n",
                        *LogicalBlock, LogicalBlocksPerZone);
                printk("LogicalBlockAddress1: 0x%x"
                        "  LogicalBlockAddress2: 0x%x\n",
                        LogicalBlockAddress1, LogicalBlockAddress2);
            }
        }
    } else {
        if ( ParityCheck1 == SMA_GOOD ) {
            *LogicalBlock = (LogicalBlockAddress1 & 0x07fe) >> 1;
            if ( *LogicalBlock < LogicalBlocksPerZone ) {
                result = SMA_HALF_OK;
            } else {
                printk("2 LogicalBlockAddress[0x%x] > 0x%x OverError!! \n",
                        *LogicalBlock, LogicalBlocksPerZone);
            }
        } else if ( ParityCheck2 == SMA_GOOD ) {
            *LogicalBlock = (LogicalBlockAddress2 & 0x07fe) >> 1;
            if ( *LogicalBlock < LogicalBlocksPerZone ) {
                result = SMA_HALF_OK;
            } else {
                printk("3 LogicalBlockAddress[0x%x] > 0x%x OverError!! \n",
                        *LogicalBlock, LogicalBlocksPerZone);
            }
        }
    }

    return result;
}

static int
CheckGetBlockAddress( struct us_data *us,
                        REDUNDANT_AREA* redundant_ptr,
                        unsigned short *LogicalBlock )
{
    int bstatus = SMA_GOOD;
    unsigned short GetLogicalBlock;
    unsigned short success_count;
    unsigned short iii;

    GetLogicalBlock = *LogicalBlock = 0xffff;

    success_count = 0;

    for ( iii=0; success_count < 3; iii++, redundant_ptr++ ) {
        if ( iii >= 6 ) {
                printk(" BlockAddressField Check Error??? (success_count=%d)\n",
                        success_count);
            bstatus = SMA_ERROR;
            break;
        }

        bstatus = CheckGetBlockAddressSub( us, redundant_ptr, &GetLogicalBlock );
        if ( ( bstatus == SMA_GOOD ) || ( bstatus == SMA_HALF_OK ) ) {
            if ( success_count == 0 ) {
                *LogicalBlock = GetLogicalBlock;
            }

            if ( GetLogicalBlock == *LogicalBlock ) {
                if ( bstatus == SMA_GOOD )
                    success_count += 2;
                if ( bstatus == SMA_HALF_OK )
                    success_count ++;
            } else {
                *LogicalBlock = GetLogicalBlock;
                success_count = 1;
            }
        } else {
            printk(" BlockAddressField Check Error!!!\n"
                        " (success_count=%d) bstatus=%d\n",
                        success_count, bstatus);
            bstatus = SMA_ERROR;
            break;
        }

    }
    if ( bstatus == SMA_HALF_OK )
        bstatus = SMA_GOOD;

    return bstatus;

}  /* CheckGetBlockAddress */

static int
CheckGetBlockAddress2( struct us_data *us,
                        REDUNDANT_AREA* redundant_ptr,
                        unsigned short* LogicalBlock )
{
    int bstatus;
    unsigned short GetLogicalBlock;

    GetLogicalBlock = *LogicalBlock = 0xffff;

    bstatus = CheckGetBlockAddressSub( us, redundant_ptr, &GetLogicalBlock );
    if ( bstatus == SMA_GOOD ) {
        *LogicalBlock = GetLogicalBlock;
    } else {
        printk(" BlockAddressField Check Fail... %04x-%04x\n",
                short_pack2(redundant_ptr->BlockAddressArea1),
                short_pack2(redundant_ptr->BlockAddressArea2) );
    }

    return bstatus;
}

/************************************************
    Make Block Allocation Table of specified Zone
************************************************/
static int
MakeBlockTableZone(struct us_data *us, unsigned int Zone)
{
    unsigned short block;
    unsigned short LogicalBlockAddress;
    unsigned short PhysicalBlock;
//    unsigned short LogicalBlocksPerZone;
//    unsigned short PhysicalBlocksPerZone;
    unsigned short TablePoint;
    int   bstatus = SMA_GOOD;
    int  loop;
    REDUNDANT_AREA* redundant;
    REDUNDANT_AREA* redundant_last;
    unsigned short LogicalBlockAddress1;
    unsigned short LogicalBlockAddress2;

    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    PhysicalBlock = (unsigned short)info->tblsz[Zone] * Zone;  // 1024*Zone
//    info->Card.Adrs = (info->tblsz[zone] * zone) << info->BitShift;

    LogicalBlockAddress = 0;

    info->ErasePoint[Zone] = 0;
    TablePoint = 0;

  {
    unsigned short *BlockTablePtr_0;
    unsigned short *BlockTablePtr;
    unsigned short *WriteEntryListPtr_0;
    unsigned short *WriteEntryListPtr;

    BlockTablePtr_0 = (unsigned short *) info->BlockTable[Zone];
    WriteEntryListPtr_0 = info->WriteEntryTableList[Zone];

    BlockTablePtr = BlockTablePtr_0;
    WriteEntryListPtr = WriteEntryListPtr_0;

    for ( block = 0; block < info->tblsz[Zone] ;
                block++, PhysicalBlock++ ) {
        info->Card.Adrs = PhysicalBlock << info->BitShift;

        while (1) {
            if ( info->MaskROM ) {
                info->BlockTable[Zone][LogicalBlockAddress] = block + (Zone * 1024);
                LogicalBlockAddress++;
                break;
            }
            for ( loop=0; loop < 5; loop++ ) {
                if ( ( bstatus = info->read_sector( us, info->Card,
                                        info->cbuf, info->rdt) )
                                == USB_STOR_TRANSPORT_GOOD )
                    break;
                sma03_ResetMedia(us);
            }
            if ( bstatus != USB_STOR_TRANSPORT_GOOD ) {
                US_DEBUGP("MakeBlockTableZone:1"
                        "  redundant read error!! : "
                        "PhysicalBlock=0x%x\n\n",
                                 PhysicalBlock);
                if ( sma03_ResultSave == (-ETIMEDOUT))
                    return (SMA_TIMEOUT);
                return (SMA_ERROR);
            }

            redundant = (PREDUNDANT_AREA) info->rdt;
            if ( redundant->BlockStatusArea != 0xff ) {
                if( Bit0Ck(info->rdt[BSA], 2) ) {  /* Error Check */
                //    printk ("MakeBlockTableZone Bit0Ck()\n");
                //    printk ("Card.Adrs = 0x%lx block = 0x%x redu = 0x%02X\n",
                //            info->Card.Adrs, block,
                //            (unsigned char)(redundant->BlockStatusArea) );
                    info->BadBlockCountPerZone[Zone] ++ ;
                    sma03_ResetMedia(us);
                    break;  // for continue;
                }
            }
            LogicalBlockAddress1 = short_pack2(redundant->BlockAddressArea1);
            LogicalBlockAddress2 = short_pack2(redundant->BlockAddressArea2);
            /* redundant->BlockAddressArea1 == redundant->BlockAddressArea2 */
            if ( (LogicalBlockAddress1 == LogicalBlockAddress2 )
                    && ( LogicalBlockAddress1 == 0xffff ) ) {
                 // = block + Zone * 1024;
                info->WriteEntryTableList[Zone][TablePoint] = PhysicalBlock;
                TablePoint++;
                break;      // for continue;
            }

            bstatus = CheckGetBlockAddress2( us,
                                redundant, &LogicalBlockAddress );
            if ( bstatus == SMA_HALF_OK )  {
                bstatus = sma03_ReadBlockRedu( us,
                        info->Card, info->rdt );
//                bstatus = sma03_ReadReduSectors( us, info->rdt,
//                                info->SectorPerBlock);
                if ( bstatus != USB_STOR_TRANSPORT_GOOD ) {
                    sma03_ResetMedia(us);
                    US_DEBUGP("MakeBlockTableZone:2  "
                        "redundant read error!! : "
                        "PhysicalBlock=0x%x\n\n", PhysicalBlock);
                    return (SMA_ERROR);
                }
                redundant = (PREDUNDANT_AREA) info->rdt;
                bstatus = CheckGetBlockAddress( us,
                                 redundant, &LogicalBlockAddress );
            }
            if ( bstatus != SMA_GOOD ) {
                info->BadBlockCountPerZone[Zone] ++ ;
                sma03_ResetMedia(us);
                break;      //continue;
            } else {
                if ( info->BlockTable[Zone][LogicalBlockAddress] == 0xffff ) {
                    info->BlockTable[Zone][LogicalBlockAddress] = PhysicalBlock;
                     // = block + zone * 1024
                } else {
                    US_DEBUGP("MakeBlockTableZone:3 adrs %ld"
                               " block:%d BAA1:%X BAA2:%X\n",
                              info->Card.Adrs, block,
                              LogicalBlockAddress1, LogicalBlockAddress2);
                    DebugDump(info->rdt, 16);  printk("\n");
                    US_DEBUGP(": LogicalBlockAddress *X Duplication Error!! :"
                           "  LogicalBlock= 0x%x\n", LogicalBlockAddress);
                    US_DEBUGP(" DUPLICATION PHYSICAL BLOCK [0x%x], InTable[0x%x]\n",
                               PhysicalBlock,
                               info->BlockTable[Zone][LogicalBlockAddress]);
//                    bstatus = ReduReadOneBlock( us, (unsigned char *)redundant );
                    bstatus = sma03_ReadBlockRedu( us,
                            info->Card, (unsigned char *)redundant );
//                    bstatus = sma03_ReadReduSectors( us,
//                                info->rdt,
//                                info->SectorPerBlock);
                    redundant = (PREDUNDANT_AREA) info->rdt;
                    redundant_last =  redundant + (info->SectorPerBlock-1);
                    LogicalBlockAddress1 = short_pack2(redundant->BlockAddressArea1);
                    LogicalBlockAddress2 = short_pack2(redundant_last->BlockAddressArea1);
                    /* redundant->BlockAddressArea1 != redundant_last->BlockAddressArea1 */
                    if ( ( LogicalBlockAddress1 != LogicalBlockAddress2 )
                            || ( bstatus != USB_STOR_TRANSPORT_GOOD ) ) {
                        US_DEBUGP(" skip this PhysicalBlock 0x%x\n", PhysicalBlock);
                        info->BadBlockCountPerZone[Zone] ++ ;
                        sma03_ResetMedia(us);
                        break;    //continue;
                    } else {
                        info->Card.Adrs =
                                (unsigned long)info->BlockTable[Zone][LogicalBlockAddress] << info->BitShift;
//                        bstatus = ReduReadOneBlock( us, (unsigned char *)redundant );
                        bstatus = sma03_ReadBlockRedu( us,
                                info->Card, (unsigned char *)redundant );
//                        bstatus = sma03_ReadReduSectors( us,
//                                        info->rdt,
//                                        info->SectorPerBlock);
                        redundant = (PREDUNDANT_AREA) info->rdt;
                        redundant_last =  redundant + (info->SectorPerBlock-1);
                        LogicalBlockAddress1 = short_pack2(redundant->BlockAddressArea1);
                        LogicalBlockAddress2 = short_pack2(redundant_last->BlockAddressArea1);
                    /* redundant->BlockAddressArea1 != redundant_last->BlockAddressArea1 */
                        if ( ( bstatus != USB_STOR_TRANSPORT_GOOD )
                                || ( LogicalBlockAddress1 != LogicalBlockAddress2 ) ) {
                            US_DEBUGP(" Set (UpdateChange) This PhysicalBlock:0x%x  "
                                        "(OLD_TablePhyBlock:0x%x)\n",
                                        PhysicalBlock,
                                        info->BlockTable[Zone][LogicalBlockAddress] );
                            info->BlockTable[Zone][LogicalBlockAddress] = PhysicalBlock;
                        } else {
                            US_DEBUGP(" skip this PhysicalBlock 0x%x\n", PhysicalBlock);
                        } /* if (bsatus.. */
                        info->BadBlockCountPerZone[Zone] ++ ;
                        sma03_ResetMedia(us);
                        break;  //continue;
                    }  /* if ( ( redundant->BlockAddressArea1 != */
                } /* if ( info->BlockTable[Zone][LogicalBlockAddress] */
            } /*  if ( bstatus...) { */
            break;
        } /* while (1) */
    }  /* for ( block = 0...tblsz[Zone] */
  }

  if (info->MaskROM == 0)  {
    for ( TablePoint = 0;
            TablePoint < info->tblsz[Zone];
            TablePoint++ ) {
        if ( info->WriteEntryTableList[Zone][TablePoint] == 0xffff )
            break;
    }
    if ( TablePoint >= info->tblsz[Zone] )
        TablePoint = 0;
    info->ErasePoint[Zone] = TablePoint;
    info->WritePoint[Zone] = 0;  /* Initial Write Entry Point */
  }

  sma03_ResetMedia(us);

  info->TableFlag[Zone] = 0;    /* Valid */

//  printk("MakeBlockTableZone: Card.Adrs = 0x%lx tblsz[0] = 0x%x\n",
//         info->Card.Adrs, info->tblsz[0]);

  return(SMA_GOOD);

} /* MakeBlockTableZone */


/************************************************
    Make Block Allocation Table ( Zone 0 )
************************************************/
static int
MakeBlockTable(struct us_data *us, unsigned int stp)
{
    unsigned int xx;
    unsigned char  Zone;
    unsigned short block;
    unsigned short LogicalBlockAddress;
    unsigned short PhysicalBlock;
    unsigned short LogicalBlocksPerZone;
    unsigned short PhysicalBlocksPerZone;
    unsigned short TablePoint;
    int   bstatus = SMA_GOOD;
    int  loop;
    REDUNDANT_AREA* redundant;
    REDUNDANT_AREA* redundant_last;
    unsigned short LogicalBlockAddress1;
    unsigned short LogicalBlockAddress2;

    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    LogicalBlocksPerZone = info->TotalLogicalBlock/info->TotalZones;
    PhysicalBlocksPerZone = info->TotalPhysicalBlock / info->TotalZones;

    for (Zone = 0; Zone < info->TotalZones; Zone++) {
        for (xx = 0; xx < LogicalBlocksPerZone; xx++) {
            info->BlockTable[Zone][xx] = 0xffff;	/* Data Clear */
            info->WriteEntryTableList[Zone][xx] = 0xffff;	/* Data Clear */
        } /* for (xx...) */
    }  /* for (Zone=0...TotalZones) */
    for (Zone = 0; Zone < info->TotalZones; Zone++) {
        for (xx = LogicalBlocksPerZone;
                xx < PhysicalBlocksPerZone; xx++)
            info->WriteEntryTableList[Zone][xx] = 0xffff;  /* Data Clear */
    }  /* for (zz = 0...info->TotalZones) */

    PhysicalBlock = stp;

    info->BadBlockCountPerZone[0] = stp;
    info->tblsz[0] = PhysicalBlocksPerZone - stp;
    info->TableFlag[0]= 0xffff;  /* Invalid */
    info->LimitBadBlock[0]
      = (info->TotalPhysicalBlock - info->TotalLogicalBlock) / info->TotalZones - stp - 1;

    for( Zone = 1; Zone < info->TotalZones; Zone++) {
        info->BadBlockCountPerZone[Zone] = 0;
        info->tblsz[Zone] = PhysicalBlocksPerZone;
        info->TableFlag[Zone] = 0xffff;  /* Data Clear */
        info->LimitBadBlock[Zone]
         = (info->TotalPhysicalBlock - info->TotalLogicalBlock) / info->TotalZones -1;
    }  /* for (zz=1...TotalZones) */

    Zone = 0;
    LogicalBlockAddress = 0;
    TablePoint = 0;    /* Pointer of unused block allocation table */

  {
    unsigned short *BlockTablePtr_0;
    unsigned short *BlockTablePtr;
    unsigned short *WriteEntryListPtr_0;
    unsigned short *WriteEntryListPtr;

    BlockTablePtr_0 = (unsigned short *) info->BlockTable[Zone];
    WriteEntryListPtr_0 = info->WriteEntryTableList[Zone];

    BlockTablePtr = BlockTablePtr_0;
    WriteEntryListPtr = WriteEntryListPtr_0;

    for ( block = stp; block < info->tblsz[Zone] ;
                block++, PhysicalBlock++ ) {

        info->Card.Adrs = (unsigned long)PhysicalBlock << info->BitShift;

        while (1) {
            if ( info->MaskROM ) {
                info->BlockTable[Zone][LogicalBlockAddress] = block;
                LogicalBlockAddress++;
                break;
            }

            for ( loop=0; loop < 5; loop++ ) {
                if ( ( bstatus = info->read_sector( us, info->Card,
                                        info->cbuf, info->rdt) )
                                == USB_STOR_TRANSPORT_GOOD )
                    break;
                sma03_ResetMedia(us);
            }
            if ( bstatus != USB_STOR_TRANSPORT_GOOD ) {
                US_DEBUGP("MakeBlockTable:1  "
                        "redundant read error!! : "
                        "PhysicalBlock=0x%x\n\n",
                                 PhysicalBlock);
                if ( sma03_ResultSave == (-ETIMEDOUT))
                    return (SMA_TIMEOUT);
                return (SMA_ERROR);
            }

            redundant = (PREDUNDANT_AREA) info->rdt;
            if ( redundant->BlockStatusArea != 0xff ) {
                if( Bit0Ck(info->rdt[BSA], 2) ) {  /* Error Check */
                    US_DEBUGP("MakeBlockTable Bit0Ck()\n");
                    US_DEBUGP("Card.Adrs = 0x%lx block = 0x%x redu = 0x%02X\n",
                            info->Card.Adrs, block,
                            (unsigned char)(redundant->BlockStatusArea) );
                    info->BadBlockCountPerZone[Zone] ++ ;
                    sma03_ResetMedia(us);
                    break;  // for continue;
                }
            }
            LogicalBlockAddress1 = short_pack2(redundant->BlockAddressArea1);
            LogicalBlockAddress2 = short_pack2(redundant->BlockAddressArea2);
            /* redundant->BlockAddressArea1 == redundant->BlockAddressArea2 */
            if ( ( LogicalBlockAddress1 == LogicalBlockAddress2 )
                        && ( LogicalBlockAddress1 == 0xffff ) ) {
                info->WriteEntryTableList[Zone][TablePoint] = block;
                TablePoint++;
                break;      // for continue;
            }

            bstatus = CheckGetBlockAddress2( us, redundant, &LogicalBlockAddress );
            if ( bstatus == USB_STOR_TRANSPORT_GOOD )  {
//                bstatus = ReduReadOneBlock( us, (unsigned char *)redundant );
                bstatus = sma03_ReadBlockRedu( us,
                        info->Card, (unsigned char *)redundant );
//                bstatus = sma03_ReadReduSectors( us,
//                                info->rdt,
//                                info->SectorPerBlock);
                if ( bstatus != USB_STOR_TRANSPORT_GOOD ) {
                    sma03_ResetMedia(us);
                    US_DEBUGP("MakeBlockTable:2  redundant read error!! : "
                           "PhysicalBlock=0x%x\n\n",
                           PhysicalBlock);
                    return (SMA_ERROR);
                }
                redundant = (PREDUNDANT_AREA) info->rdt;
                bstatus = CheckGetBlockAddress( us, redundant, &LogicalBlockAddress );
            }

            if ( bstatus != USB_STOR_TRANSPORT_GOOD ) {
                US_DEBUGP("MakeBlockTable:3 "
                        "adrs %ld block:%d BAA1:%X BAA2:%X\n",
                        info->Card.Adrs, block,
                        LogicalBlockAddress1, LogicalBlockAddress2);
                        DebugDump(info->rdt, 16);  printk("\n");
                info->BadBlockCountPerZone[Zone] ++ ;
                sma03_ResetMedia(us);
                break;      //continue;
            } else {
                if ( info->BlockTable[Zone][LogicalBlockAddress] == 0xffff ) {
                    info->BlockTable[Zone][LogicalBlockAddress] = block;
                } else {
                    US_DEBUGP("MakeBlockTable:4 "
                              "adrs %ld block:%d BAA1:%X BAA2:%X\n",
                              info->Card.Adrs, block,
                              LogicalBlockAddress1, LogicalBlockAddress2);
                              DebugDump(info->rdt, 16);  printk("\n");
                    US_DEBUGP(": LogicalBlockAddress *X Duplicate Error!! :"
                               "  LogicalBlock= 0x%x\n", LogicalBlockAddress);
                    US_DEBUGP(" Duplicate PHYSICAL BLOCK [0x%x], InTable[0x%x]\n",
                              PhysicalBlock,
                              info->BlockTable[Zone][LogicalBlockAddress]);

//                    bstatus = ReduReadOneBlock( us, (unsigned char *)redundant );
//                    bstatus = sma03_ReadBlockRedu( us,
//                            info->Card, (unsigned char *)redundant );
                    bstatus = sma03_ReadReduSectors( us,
                                info->rdt,
                                info->SectorPerBlock);
                    redundant = (PREDUNDANT_AREA) info->rdt;
                    redundant_last =  redundant + (info->SectorPerBlock-1);
                    LogicalBlockAddress1 = short_pack2(redundant->BlockAddressArea1);
                    LogicalBlockAddress2 = short_pack2(redundant_last->BlockAddressArea1);
                    /* redundant->BlockAddressArea1 != redundant_last->BlockAddressArea1 */
                    if ( ( LogicalBlockAddress1 != LogicalBlockAddress2 )
                            || ( bstatus != USB_STOR_TRANSPORT_GOOD ) ) {
                        US_DEBUGP(" skip this PhysicalBlock 0x%x\n", PhysicalBlock);
                        info->BadBlockCountPerZone[Zone] ++ ;
                        sma03_ResetMedia(us);
                        break;    //continue;
                    } else {
                        info->Card.Adrs =
                                (unsigned long)info->BlockTable[Zone][LogicalBlockAddress] << info->BitShift;
//                        bstatus = ReduReadOneBlock( us, (unsigned char *)redundant );
//                        bstatus = sma03_ReadBlockRedu( us,
//                                info->Card, (unsigned char *)redundant );
                        bstatus = sma03_ReadReduSectors( us,
                                        info->rdt,
                                        info->SectorPerBlock);
                        LogicalBlockAddress1 = short_pack2(redundant->BlockAddressArea1);
                        LogicalBlockAddress2 = short_pack2(redundant_last->BlockAddressArea1);
                        redundant = (PREDUNDANT_AREA) info->rdt;
                        redundant_last =  redundant + (info->SectorPerBlock-1);
                    /* redundant->BlockAddressArea1 != redundant_last->BlockAddressArea1 */
                        if ( ( bstatus != USB_STOR_TRANSPORT_GOOD )
                                || ( LogicalBlockAddress1 != LogicalBlockAddress2 ) ) {
                            US_DEBUGP(" Set (UpdateChange) This PhysicalBlock:0x%x  "
                                        "(OLD_TablePhyBlock:0x%x)\n",
                                        PhysicalBlock,
                                        info->BlockTable[Zone][LogicalBlockAddress] );
                            info->BlockTable[Zone][LogicalBlockAddress] = block;
                        } else {
                            US_DEBUGP(" skip this PhysicalBlock 0x%x\n", PhysicalBlock);
                        } /* if (bsatus.. */
                        info->BadBlockCountPerZone[Zone] ++ ;
                        sma03_ResetMedia(us);
                        break;  //continue;
                    }  /* if ( ( redundant->BlockAddressArea1 != */
                } /* if ( info->BlockTable[Zone][LogicalBlockAddress] */
            } /*  if ( bstatus...) { */
            break;
        } /* while (1) */
    }  /* for ( block = stp...tblsz[Zone] */
  }

  if (info->ReadOnly == 0)  {
    Zone = 0;
    for ( TablePoint = 0;
            TablePoint < info->tblsz[Zone];
            TablePoint++ ) {
        if ( info->WriteEntryTableList[Zone][TablePoint] == 0xffff )
            break;
    }
    if ( TablePoint >= info->tblsz[Zone] )
        TablePoint = 0;
        info->ErasePoint[Zone] = TablePoint;
        info->WritePoint[Zone] = 0;  /* Initial Write Entry Point */
  }

//  printk("MakeBlockTable: BlockTable[0][0]: %d\n",
//   info->BlockTable[0][0] );

  sma03_ResetMedia(us);

  info->TableFlag[Zone] = 0;  /* Valid */

  return(SMA_GOOD);
}  /* MakeBlockTable */

static void
ConvertMediaBuffer512to256( struct us_data *us,
                        unsigned short  sectors )
{
    int         iii;
    unsigned char *ptr_dest;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);
    PBLOCKREAD_MEMORYMAP512 mapptr512;
    PBLOCKREAD_MEMORYMAP256 mapptr256;

    if ( info->pagesize == Page512M ) {
        return;
    }

    ptr_dest = info->MyMediaBuf;
    mapptr512 = (PBLOCKREAD_MEMORYMAP512)ptr_dest;
    mapptr256 = (PBLOCKREAD_MEMORYMAP256)ptr_dest;


  // 256 Byte/Page      adding64
  // 576 Byte/Sector    40 Byte/Sector
  //mapptr256 = (PBLOCKREAD_MEMORYMAP256)(DDBPtr->MediaBufSLR+640);
  //mapptr512++;        // [576]
    mapptr256 = (PBLOCKREAD_MEMORYMAP256)(info->MyMediaBuf+640*(sectors-1));
    mapptr512 = (PBLOCKREAD_MEMORYMAP512)(info->MyMediaBuf+576*(sectors-1));

    for ( iii=1; iii < sectors; iii++ ) {
        memcpy( (unsigned char *)mapptr256, (unsigned char *)mapptr512, 512+16 );
        mapptr512--;    //+=576
        mapptr256--;    //+=640
    }

    mapptr256 = (PBLOCKREAD_MEMORYMAP256)ptr_dest;
    for ( iii=0; iii < sectors; iii++ ) {
        // 256 Byte/Page
        // Data256+Red8+Padding56+Data256+Red8+Padding56 = 640

        memcpy( (unsigned char *)(mapptr256->DataArea1),
                (unsigned char *)(&(mapptr256->RedundantArea0)), 256+16 );

        memcpy( (unsigned char *)(&(mapptr256->RedundantArea0)),
                (unsigned char *)(&(mapptr256->RedundantArea1)), 8 );
        memcpy( (unsigned char *)(&(mapptr256->RedundantArea1)),
                (unsigned char *)(mapptr256->Padding1), 8 );
        mapptr256++;    // +=640
    }

    return;
}

/***********************************************
        Write Redundant New Data
***********************************************/
static void
SetRedundantData( struct us_data *us,
                PBLOCKREAD_MEMORYMAP512 MediaBufPtr,
                unsigned short  start_sector,
                unsigned short  RWSectors )
{
    int iii;
    unsigned short  BlockAddressFieldData;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    US_DEBUGP(" SetRedundantData... \n");

    BlockAddressFieldData
         = LogicalBlockToBlockAddressField( us, info->ac_block );

    US_DEBUGP("   zone=[%d]: set LogicalBlock=0x%x => "
        "BlockAddressField=0x%x\n",
        info->ac_zone, info->ac_block, BlockAddressFieldData);

    for ( iii=0; iii < RWSectors; iii++ ) {
        memset( (unsigned char *)MediaBufPtr->DataArea, 0xff, 512 );
        memset( (unsigned char *)&MediaBufPtr->RedundantArea, 0xff, 64 );
        *(unsigned short *)(MediaBufPtr->RedundantArea.BlockAddressArea1)
             = BlockAddressFieldData;
        *(unsigned short *)(MediaBufPtr->RedundantArea.BlockAddressArea2)
             = BlockAddressFieldData;

        MediaBufPtr++;  // += 576
    }

    return;
}  /* SetRedundantData */

static void
CopySectorDataFromSgList(struct us_data *us,
             PBLOCKREAD_MEMORYMAP512 *ToBuf)
{
    int  cpy_len;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);
    PBLOCKREAD_MEMORYMAP512  ptr = *ToBuf;

    /*
     * No use sg list, or use only this sg list
     */

    if ( !(info->use_sg) || info->sg_length >= 512) {
        memcpy(ptr->DataArea, info->sg_ptr, 512);
        info->sg_ptr += 512;
        info->sg_length -= 512;
        /* Set Redundant */
        memset((unsigned char *)(&ptr->RedundantArea), 0xff, 64 );
        ptr++;
        *ToBuf = ptr;
        return;
    }

    /*
     * copy to this sg list and next sg list
     */

    /* copy to this sg (sg_length) */
    memcpy(ptr->DataArea, info->sg_ptr, info->sg_length);
    info->sg_no++;   /* next sg buffer */
    info->sg_ptr = info->sg[info->sg_no].address;
    /* copy (512 - sg_length) */
    cpy_len = 512 - info->sg_length;
    memcpy(ptr->DataArea + info->sg_length,
                info->sg_ptr, cpy_len );
    /* Set Redundant */
    memset((unsigned char *)(&ptr->RedundantArea), 0xff, 64 );
    /* next pointer */
    info->sg_length = info->sg[info->sg_no].length - cpy_len;
    info->sg_ptr += cpy_len;
    ptr++;
//  printk("--- after %p-%d\n", info->sg_ptr, info->sg_length);
    *ToBuf = ptr;
    return;
}  /* CopySectorDataFromSgList */

/*
 *
 */
static void
SetWriteNewData( struct us_data *us,
                unsigned char  *buf,
                unsigned short  RWSectors )
{
    int iii;
    unsigned char      *dap;
    unsigned short     *bap1, *bap2;
    PBLOCKREAD_MEMORYMAP512 MediaBufPtr;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    US_DEBUGP(" SetWriteNewData... \n");

    MediaBufPtr = (PBLOCKREAD_MEMORYMAP512)buf;

//    printk("SetWriteNewData\nMediaBufPtr: %p",  MediaBufPtr);
//    printk(" DataArea: %p",  MediaBufPtr->DataArea);

    for ( iii=0; iii < RWSectors; iii++ ) {
//        memcpy( (unsigned char *)MediaBufPtr->DataArea,
//                (unsigned char *)info->trspt, 512 );
//        memset( (unsigned char *)&MediaBufPtr->RedundantArea, 0xff, 64 );

        dap = (unsigned char *)MediaBufPtr->DataArea;
        bap1 = (unsigned short *)(MediaBufPtr->RedundantArea.BlockAddressArea1);
        bap2 = (unsigned short *)(MediaBufPtr->RedundantArea.BlockAddressArea2);
        CopySectorDataFromSgList(us, &MediaBufPtr);

        *bap1 = *bap2
              = LogicalBlockToBlockAddressField( us, info->ac_block );

        EccCodeHand( us, dap, WRITE_DIRECTION );

//        info->trspt += 512;
//        MediaBufPtr ++; // += 576

    }

    return;
}  /* SetWriteNewData */

/*
 *
 *     Now Block Data to New Block Copy
 *
 */
static int
SetOrginalData( struct us_data *us,
                        PBLOCKREAD_MEMORYMAP512 MediaBufPtr,
                        unsigned short start_sector,
                        unsigned short RWSectors )
{
    unsigned short  page;
    unsigned short  OriginalBlock;
    int             status;
    int             ErrorFlag = FALSE;
    unsigned short  iii;
    unsigned short  BlockAddressFieldData;
    PBLOCKREAD_MEMORYMAP512 ReadMediaBufPtr;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    US_DEBUGP(" SetOrginalData...\n");


    page = start_sector * info->PagePerSector;
    OriginalBlock = info->BlockTable[info->ac_zone][info->ac_block];

  // ******* Read Original Block Data  *******
    info->Card.Adrs = ((unsigned long)OriginalBlock << info->BitShift) + page;

/*
    if ( ( info->FWInfo.CodeLevel & CLEVEL_READBLOCK_SUPPORT )
            == CLEVEL_READBLOCK_SUPPORT ) {
        status = sma03_BlockRead( us, MediaBufPtr, RWSectors );
    } else {
        status = sma03_ReadSectors( us, MediaBufPtr, RWSectors );
    }
*/
    status = info->read_block( us, MediaBufPtr, RWSectors );   /* 2001-11-26 */
    if ( status ) {
        printk(" ERROR!! BlockRead( for WriteBlock ) ReadError!!\n");
        return(SMA_ERROR);
    }

  // for ECC CHECK
    ErrorFlag = FALSE;
    ReadMediaBufPtr = MediaBufPtr;
    for ( iii = 0; iii < RWSectors; iii++ ) {
        status = EccCodeHand( us, ReadMediaBufPtr->DataArea, READ_DIRECTION );
        if ( status ) { // ECC check Error!!
            ErrorFlag = TRUE;
            printk(" [DANGER! DANGER!]  ECC ERROR !! "
                "Block=0x%x\n [DANGER!!] KOKODE OKITARA DOUSHIYOUMO NAI!!\n",OriginalBlock);
            memset(&ReadMediaBufPtr->RedundantArea, 0xff, 64);
            BlockAddressFieldData = LogicalBlockToBlockAddressField( us, info->ac_block );
            *(unsigned short *)(ReadMediaBufPtr->RedundantArea.BlockAddressArea1)
             = BlockAddressFieldData;
            *(unsigned short *)(ReadMediaBufPtr->RedundantArea.BlockAddressArea2)
             = BlockAddressFieldData;
            EccCodeHand( us, ReadMediaBufPtr->DataArea, WRITE_DIRECTION );

            ReadMediaBufPtr->RedundantArea.DataStatusArea = 0x00;

        }
        ReadMediaBufPtr ++;
    }

    if ( ErrorFlag == TRUE ) {
        BadBlockSet( us, OriginalBlock );
        //status = NG;
        status = SMA_ECC_MARKED_F0; // = -1
    } else {
        status = SMA_GOOD;
    }
    return status;
}  /* SetOrginalData */

static int
VerifySub( struct us_data *us,
           PBLOCKREAD_MEMORYMAP512 CompareMediaBufPtr,
           unsigned short  ReadPhysicalBlock,
           unsigned short  Sectors )
{
    unsigned char  *VerifyBufSLR;
    unsigned long  SLRBufferSize;
    unsigned long  PageByteLength;
    unsigned long  len;
    PBLOCKREAD_MEMORYMAP512 MediaBufPtr;
    unsigned short iii;
    int    bstatus;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    US_DEBUGP("VerifyCheck... PhysicalBlock=0x%x\n",ReadPhysicalBlock );

    if ( info->pagesize != Page512M ) {
        US_DEBUGP("VerifyCheck... NotExecute 256Byte/Page Media!!!\n" );
        return SMA_NOT_EXECUTE;
    }

    if ( ( info->FWInfo.CodeLevel & (CLEVEL_READBLOCK_SUPPORT | CLEVEL_READBLOCK_SUPPORT) )
            != (CLEVEL_READBLOCK_SUPPORT | CLEVEL_READBLOCK_SUPPORT) ) {
        US_DEBUGP("VerifyCheck... NotExecute SMA03 Mode!!!\n" );
        return SMA_NOT_EXECUTE;
    }

    SLRBufferSize  = Sectors * (512+64);
    PageByteLength = 528;

    VerifyBufSLR = kmalloc(SLRBufferSize, GFP_KERNEL);
    if ( VerifyBufSLR == NULL ) {
          // Memory Allocate Error!!
        US_DEBUGP(" Verify kmalloc(0x%lx) fail...\n", SLRBufferSize );
        return SMA_NOT_EXECUTE;
    }

    MediaBufPtr = (PBLOCKREAD_MEMORYMAP512)VerifyBufSLR;
      // ******* Read Original Block Data  *******
    info->Card.Adrs
     = (unsigned long)ReadPhysicalBlock << info->BitShift;
//    bstatus = sma03_BlockRead( us, (PBLOCKREAD_MEMORYMAP512)VerifyBufSLR, Sectors );
    bstatus = info->read_block( us, MediaBufPtr, Sectors );   /* 2001-11-26 */
    if ( bstatus != SMA_GOOD ) {
        // ReadError !!!
         US_DEBUGP("\n Verify ReadError!!... PhysicalBlock=0x%x\n",
                ReadPhysicalBlock );
         bstatus = SMA_NOT_EXECUTE;
    } else {
        bstatus = SMA_GOOD;
        for ( iii=0; iii < Sectors; iii++ ) {
            bstatus = EccCodeHand( us, MediaBufPtr->DataArea, READ_DIRECTION );
            if ( bstatus ) { // ECC check Error!!
                bstatus = SMA_VERIFY_CHECK_ERROR;	// ==NG
                US_DEBUGP("\n [!!! DANGER!!!]  Verify : "
                       "ReadMedia ECC Error !!! ***** \n\n");
                US_DEBUGP("   ReadPhysicalBlock= 0x%x,  sector=%d\n",
                         ReadPhysicalBlock, iii );
                break;
            }

            len = memcmp( (unsigned char *) CompareMediaBufPtr,
                        (unsigned char *) MediaBufPtr,
                        (unsigned long) PageByteLength );
            if ( len != PageByteLength ) {
              // Verify : CompareError!!
                bstatus = SMA_VERIFY_CHECK_ERROR;  // ==NG
                printk("\n [!!! DANGER!!!]  VerifyError !!! ***** \n");
                printk("   ReadPhysicalBlock= 0x%x,  sector=%d\n",
                ReadPhysicalBlock, iii );
                printk("   CompareMediaBufPtr=0x%p : "
                       "MediaData=0x%p \n"
                       "   failData Offset = [0x%lx]\n\n",
                        CompareMediaBufPtr,
                        MediaBufPtr, len );
                break;
            }
            CompareMediaBufPtr++;
            MediaBufPtr++;
        }   /* for ( iii=0...Sectors ) */
    }
    kfree( VerifyBufSLR );

    return bstatus;

}  /* VerifySub */


/*
 *  Setup Block Allocation Table
 */
int sma03_read_map(struct us_data *us)
{
    int  result;
    unsigned short  CisBlock;
//    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    US_DEBUGP("sma03_read_map: start\n");
    GenCPtbl(us);       /* Generate C Parity ECC Record to ecctbl[] */
    if (( result = CisSearchAndCheck( us, &CisBlock)) != SMA_GOOD)
        return USB_STOR_TRANSPORT_ERROR;
    US_DEBUGP("sma03_read_map: CisSearchAndCheck:OK  CisBlock:%d\n", CisBlock);

    result = HiddRd(us);         /* Read Hidden Block */
    US_DEBUGP("sma03_read_map: HiddRd: %d\n", result);

    if (MakeBlockTable(us, CisBlock+1) != SMA_GOOD)  /* Make Block Allocation Table */
        return USB_STOR_TRANSPORT_ERROR;
    US_DEBUGP("sma03_read_map: MakeBlockTable: OK\n");

    return USB_STOR_TRANSPORT_GOOD;
}  /* sma03_read_map */


void sma03_card_info_destructor(void *extra) {
//      struct sma03_card_info *info = (struct sma03_card_info *)extra;

	if (!extra)
		return;
/*
	if (info->lba_to_pba)
		kfree(info->lba_to_pba);
	if (info->pba_to_lba)
		kfree(info->pba_to_lba);
*/
}  /* sma03_card_info_destructor */

static int sma03_rw_sector(Scsi_Cmnd *srb, struct us_data *us)
{
 printk("sma03_rw_sector: \n");
            return USB_STOR_TRANSPORT_ERROR;

/*
        result = sma03_ReadPage( us, 0, page,
                         srb->request_buffer, srb->use_sg);
*/
}
/*
 * SCSI Command Emulation for the RATOC SMA03
 *
 * Test Unit Ready (00)
 */
static int
sma03_TestUnitReady(Scsi_Cmnd *srb, struct us_data *us)
{
    int     result = 0;
    int     loop;
    unsigned char buf[8];
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    current_sense_data = NULL;
    buf[0] = 0;

    srb->result = 0;

    for (loop = 0; loop < 5; loop++) {
        result = sma03_send_control(us,
                            usb_rcvctrlpipe(us->pusb_dev, 0),
                            SMA03_IN_DATA,
                            USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                            0, SMA03PORT_USB_STAT,
                            buf, 1);
        if (result == USB_STOR_TRANSPORT_GOOD)
            break;
    }
    if (result != USB_STOR_TRANSPORT_GOOD)  {
        printk("sma03_TestUnitReady: result:%d\n", result);
        current_sense_data = notReady;
        memcpy(srb->sense_buffer, current_sense_data, 18);
        info->ComputeGeomFlag = 0;
        srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
        return result;
    }

    if ( buf[0] == SMA03_STAT_INIT ) { /* Initilizing */
        mdelay(500L);  /* Wait for initilizing H/W */
        current_sense_data = mediumNotPresent;
        memcpy(srb->sense_buffer, current_sense_data, 18);
        srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
//        srb->allowed = 1000;    // Max retry count 1000
        return USB_STOR_TRANSPORT_GOOD;
    }

    if ( !(buf[0] & SMA03_STAT_CENB) ) {   /* No Media */
        current_sense_data = mediumNotPresent;
        /* set the result so the higher layers expect this data */
        memcpy(srb->sense_buffer, current_sense_data, 18);
        info->ComputeGeomFlag = 0;
        srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
        return USB_STOR_TRANSPORT_GOOD;
    } else
    if ( buf[0] & SMA03_STAT_STCHG ) {  /* Media changed */
        result = sma03_send_control(us,
                    usb_sndctrlpipe(us->pusb_dev, 0),
                    SMA03_OUT_DATA,
                    USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                    SMA03_MODE_STANDBY, SMA03PORT_USB_MODE,
                    buf, 0);
//        current_sense_data = notReady;
        current_sense_data = mediumNotPresent;
        memcpy(srb->sense_buffer, current_sense_data, 18);
        info->ComputeGeomFlag = 0;
//        srb->device->changed = 1;
        srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
        return USB_STOR_TRANSPORT_GOOD;
    }

    if (info->ComputeGeomFlag == 0) {
        /*
         * Set notReady till changing geometry parameters.
         */
        current_sense_data = notReady;
        memcpy(srb->sense_buffer, current_sense_data, 18);
        srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
        return USB_STOR_TRANSPORT_GOOD;
    }
   if (srb->result)
   printk("sma03_TestUnitReady: srb->result:%04X sense:%02X\n",
            srb->result , srb->sense_buffer[12]);

    return USB_STOR_TRANSPORT_GOOD;
}  /* sma03_TestUnitReady */

static void
CopyToSgList(struct us_data *us,
             PBLOCKREAD_MEMORYMAP512 *FromBuf,
                int  use_sg)
{
    int  cpy_len;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);
    PBLOCKREAD_MEMORYMAP512  ptr = *FromBuf;

    /*
     * No use sg list, or use only this sg list
     */

    if ( !use_sg || info->sg_length >= 512) {
        memcpy(info->sg_ptr, ptr->DataArea, 512);
        info->sg_ptr += 512;
        info->sg_length -= 512;
        ptr++;
        *FromBuf = ptr;
        return;
    }

    /*
     * copy to this sg list and next sg list
     */

    /* copy to this sg (sg_length) */
    memcpy(info->sg_ptr, ptr->DataArea, info->sg_length);
    info->sg_no++;   /* next sg buffer */
    info->sg_ptr = info->sg[info->sg_no].address;
    /* copy (512 - sg_length) */
    cpy_len = 512 - info->sg_length;
    memcpy(info->sg_ptr,
           ptr->DataArea + info->sg_length, cpy_len );
    /* next pointer */
    info->sg_length = info->sg[info->sg_no].length - cpy_len;
    info->sg_ptr += cpy_len;
    ptr++;
//  printk("--- after %p-%d\n", info->sg_ptr, info->sg_length);
    *FromBuf = ptr;
    return;
}

static void
SetToSgList(struct us_data *us,
             unsigned long bytecount,
             int  use_sg)
{
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    if ( !use_sg ) {  /* No scatter gather */
        memset(info->sg_ptr, 0xff, bytecount);
        info->sg_ptr += bytecount;
        return;
    }

    /*
     * Use scatter gather
     */

    while (1) {
        /* set to this sg (bytecount) */
        if (info->sg_length >= bytecount) {
            memset(info->sg_ptr, 0xff, bytecount);
            info->sg_ptr += bytecount;
            info->sg_length -= bytecount;
            return;
        }
        /* set to this sg (sg_length) */
        memset(info->sg_ptr, 0xff, info->sg_length);
        bytecount -= info->sg_length;  /* remain */
        info->sg_no++;   /* next sg buffer */
        info->sg_ptr = info->sg[info->sg_no].address;
        info->sg_length = info->sg[info->sg_no].length;
    }

}
static int
sma03_RequestSense(Scsi_Cmnd *srb, struct us_data *us)
{
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    US_DEBUGP("sma03_RequestSense: \n");

    memcpy(srb->sense_buffer, current_sense_data, 18);
    srb->result = 0;
    current_sense_data = noSense;
    return USB_STOR_TRANSPORT_GOOD;
}
/*
 * SCSI Command Emulation for the RATOC SMA03
 *
 * Read (28)
 */
static int
sma03_Read(Scsi_Cmnd *srb, struct us_data *us,
                unsigned int EccVerify )  /* 1:ReadVerify-on 0:off */
{

    unsigned long ReadBufferSize;
//    unsigned char  *XferBufferPtr;
    PBLOCKREAD_MEMORYMAP512  ReadMediaBufPtr = NULL;
    PBLOCKREAD_MEMORYMAP512  ptr = NULL;
    unsigned short ReadSectorCount = 0;
    unsigned short ReadSectors;
    unsigned short Block;
    unsigned short EndBlock;
    unsigned long NowSector;
    unsigned long EndSector;
    int     PhysicalBlock;
//    int     cpy_len;
    int     bstatus;
    register unsigned int iii;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    if ( (bstatus = sma03_TestUnitReady(srb, us)) )
        return bstatus;
    if ( info->ComputeGeomFlag == 0 ) {

        /* Media changed */
        US_DEBUGP("Media changed\n");
            srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
            return USB_STOR_TRANSPORT_ERROR;
    }

    if (srb->result) {
        if (srb->sense_buffer[12] != 0x28)
            return bstatus;
        /* Media changed */
        US_DEBUGP("Media changed\n");
        return USB_STOR_TRANSPORT_ERROR;
    }

    info->ac_cunt = short_pack(srb->cmnd[8], srb->cmnd[7]); /* Number of sectors to read */
    info->ac_strt = short_pack(srb->cmnd[3], srb->cmnd[2]);
    info->ac_strt <<= 16;
    info->ac_strt |= short_pack(srb->cmnd[5], srb->cmnd[4]); /* Start Sector to read */
    info->use_sg = srb->use_sg;
    if ( !info->use_sg ) {
        info->trspt = srb->request_buffer;   /* Read buffer pointer */
        info->sg_length = srb->request_bufflen;
        info->sg_ptr = info->trspt;
    } else {
        info->sg = (struct scatterlist *)(srb->request_buffer);
        info->sg_no = 0;
        info->trspt = info->sg[0].address;
        info->sg_length = info->sg[0].length;
        info->sg_ptr = info->trspt;
    }
//  printk("SectorRead : use_sg: %d\n", srb->use_sg);
//  printk(" start sector = %08lx, count = %08lx\n bufflen = %04x\n",
//        info->ac_strt, info->ac_cunt, srb->request_bufflen);

    NowSector = info->ac_strt;             /* sector no. of current read */
    EndSector = NowSector + info->ac_cunt -1;
    Block = (unsigned short)(NowSector / info->SectorPerBlock);  /* block no. of current read */
    EndBlock = (unsigned short)(EndSector / info->SectorPerBlock);

    info->secnm = (unsigned short)(info->ac_strt % info->SectorPerBlock);
         /* Start Read Sector in block */
    info->ac_strt = info->ac_strt / info->SectorPerBlock;
         /* Block Allocation Table Point */
    info->ac_zone = (unsigned short)(info->ac_strt / MAX_BLOCK_PER_ZONE_LOGICAL);
    info->ac_block = (unsigned short)(info->ac_strt % MAX_BLOCK_PER_ZONE_LOGICAL);


    /* Set ECC Check routin entry */
    if ( EccVerify ) {
        info->check_ecc = EccCodeHand;  /* ReadVerify ECC from data */
    } else {
        info->check_ecc = EccCodeCheck; /* Check ECC-code only */
    }

    if ( info->TableFlag[info->ac_zone] == 0xffff ) {   /* Invalid */
     if (info->ac_zone == 0)
      printk("sma03_Read:1 MakeBlockTableZone() : %d\n", info->ac_zone);
        if ( MakeBlockTableZone(us, info->ac_zone) )  { /* Make Block Allocation Table */
                        /*
                         *   Must set Sense data
                         */

            srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
            return USB_STOR_TRANSPORT_ERROR;
        }
    }

    PhysicalBlock = info->BlockTable[info->ac_zone][info->ac_block];

    if (PhysicalBlock != 0x0000ffff) {
        info->Card.Adrs = PhysicalBlock << info->BitShift;
        info->Card.b.adr0 += ((unsigned char)info->secnm * info->PagePerSector);
    }

//    printk("Read ZONE = %d, Block = %04x\n",
//        info->ac_zone, info->ac_block);
//    printk(" PhysicalBlock = %08x\n", PhysicalBlock );
//    printk(" SectorInBlock = %04x, SecCount = %08lx\n",
//        info->secnm, info->ac_cunt);

    if ( info->ac_cunt == 1) {
//        printk("SingleSector %ld\n", info->Card.Adrs);
        if ( (bstatus = info->read_sector(us,
                                info->Card,
                                info->trspt,
                                info->rdt))
                        != USB_STOR_TRANSPORT_GOOD) {
            srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
            return bstatus;
        }
        if (info->MaskROM == 0)  {
           if ( (bstatus = EccCodeHandSingle( us,
                           info->trspt, info->rdt, READ_DIRECTION) ) )  {
                        /* ECC chaeck error */
/*
  printk(" ");   DebugDump((info->trspt), 16);  printk("\n");
  printk(" ");   DebugDump((info->trspt + 0x10), 16);  printk("\n");
  printk(" ");   DebugDump((info->trspt + 0x20), 16);  printk("\n");
  printk(" ");   DebugDump((info->trspt + 0x30), 16);  printk("\n");
  printk("Redu:\n");
  printk(" ");   DebugDump((info->rdt), 16);  printk("\n");
*/
               srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
               return bstatus;
           }
        }
        srb->result = 0x0000;
        return USB_STOR_TRANSPORT_GOOD;
    }

    for ( ; Block < (EndBlock+1); Block++) {
        if ( Block == EndBlock )   /* Last Block ? */
            ReadSectors = (unsigned short)(EndSector - NowSector + 1);
        else
            ReadSectors
                = (unsigned short)(info->SectorPerBlock
                         - (NowSector % info->SectorPerBlock) );

        if ( PhysicalBlock != 0x0000ffff ) {
            if (info->pagesize == Page512M )
                ReadBufferSize = ReadSectors * (512 + 64 );
            else
                ReadBufferSize = ReadSectors * (512 + 64 + 64);

            if (((unsigned char *)ReadMediaBufPtr
                  = kmalloc( ReadBufferSize, GFP_KERNEL)) == NULL) {
                printk("sma03_Read: can't allocate ReadMediaBuf\n"
                        " size=%ld\n", ReadBufferSize);
                return USB_STOR_TRANSPORT_ERROR;
            }

/*
          if ( (bstatus = sma03_BlockRead(us,
                            ReadMediaBufPtr,
                            ReadSectors)
                ) != USB_STOR_TRANSPORT_GOOD)  {
*/
          bstatus = info->read_block( us, ReadMediaBufPtr, ReadSectors );   /* 2001-11-26 */
          if (bstatus != USB_STOR_TRANSPORT_GOOD)  {
                srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                kfree (ReadMediaBufPtr);
                return bstatus;
            }

            ptr = ReadMediaBufPtr;
            for (iii = 0; iii < ReadSectors; iii++) {
                if (info->MaskROM == 0)  {
                    if ( (bstatus = EccCodeHand( us,
                                    ptr->DataArea,
                                    READ_DIRECTION) ) )  {
  printk("page: %ld ptr:%p\n", info->Card.Adrs + iii, ptr);
  printk(" ");   DebugDump((ptr->DataArea), 16);  printk("\n");
  printk(" ");   DebugDump((ptr->DataArea + 0x10), 16);  printk("\n");
  printk(" ");   DebugDump((ptr->DataArea + 0x20), 16);  printk("\n");
  printk(" ");   DebugDump((ptr->DataArea + 0x30), 16);  printk("\n");
  printk("Redu:\n");
  printk(" ");   DebugDump(((unsigned char *)&(ptr->RedundantArea)), 16);  printk("\n");

                        /* ECC chaeck error */
                        srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                        kfree (ReadMediaBufPtr);
                        return bstatus;
                    }
                }  /* info->ReadOnly == 0 */
                /*
                 * Copy data to srb->request_buffer
                 * Set to next buffer (ptr++)
                 */
                CopyToSgList(us, &ptr, srb->use_sg);
            }  /* iii = 0...ReadSectors */
            if (ReadMediaBufPtr != NULL)
                kfree (ReadMediaBufPtr);
         /* PhysicalBlock != 0xffff */
        } else {
        /* PhysicalBlock == 0xffff */
            SetToSgList(us,
                        (unsigned long)(ReadSectors) * 512L,
                        srb->use_sg);
        }
        NowSector += ReadSectors;
        ReadSectorCount += ReadSectors;

        if (Block == EndBlock)
            break;

        info->ac_block++;  /* next block in zone */
        if (info->ac_block >= MAX_BLOCK_PER_ZONE_LOGICAL) {
            info->ac_block = 0; /* block in zone = 0 */
            info->ac_zone++;     /* next zone */
            if ( info->TableFlag[info->ac_zone] == 0xffff ) {
     if (info->ac_zone == 0)
      printk("sma03_Read:2 MakeBlockTableZone() : %d\n", info->ac_zone);
                if ( (bstatus = MakeBlockTableZone(us,
                                        info->ac_zone) ) ) {
                    srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                    return bstatus;
                }
            }
        }  /* info->ac_block >= MAX_BLOCK_PER_ZONE_LOGICAL */

        PhysicalBlock = info->BlockTable[info->ac_zone][info->ac_block];
        if (PhysicalBlock != 0xffff)
            info->Card.Adrs = PhysicalBlock << info->BitShift;
    }  /* for (Block= ...EndBlock+1; */
//  printk("\n"); DebugDump( ((unsigned char *)(srb->request_buffer)), 16);
//  printk("\n"); DebugDump( ((unsigned char *)(srb->request_buffer)+0x1B0 ), 16);
//  printk("\n");

    current_sense_data = noSense;
    srb->result = 0x0000;

    return USB_STOR_TRANSPORT_GOOD;

}  /* sma03_Read */

/*
 * SCSI Command Emulation for the RATOC SMA03
 *
 * Write (2A)
 */
static int
sma03_Write(Scsi_Cmnd *srb, struct us_data *us,
                unsigned int EccVerify )  /* 1:EccVerify-on 0:off */
{
    unsigned short  RW_SectorCount;
    int             erase_flag;
    unsigned long   w_count;
    unsigned long   remain_count;
    unsigned short  oldBlock = 0;
    unsigned short  NewBlock;
    unsigned short  MediaStatus;
    int             WriteFailCounter;

    int             VerifyFailCounter;
    int             UsbOutFailCounter;
    int             bstatus;
    unsigned long   WriteBufferSize;
    PBLOCKREAD_MEMORYMAP512  MediaBufPtr;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    if ( (bstatus = sma03_TestUnitReady(srb, us)) )
        return bstatus;

    info->ac_cunt = short_pack(srb->cmnd[8], srb->cmnd[7]); /* Number of sectors to read */
    info->ac_strt = short_pack(srb->cmnd[3], srb->cmnd[2]);
    info->ac_strt <<= 16;
    info->ac_strt |= short_pack(srb->cmnd[5], srb->cmnd[4]); /* Start Sector to read */
    info->use_sg = srb->use_sg;
    if ( !info->use_sg ) {
        info->trspt = srb->request_buffer;   /* Write buffer pointer */
        info->sg_length = srb->request_bufflen;
        info->sg_ptr = info->trspt;
    } else {
        info->sg = (struct scatterlist *)(srb->request_buffer);
        info->sg_no = 0;
        info->trspt = info->sg[0].address;
        info->sg_length = info->sg[0].length;
        info->sg_ptr = info->trspt;
    }

    US_DEBUGP("SectorWrite : start sector = %08lx, count = %08lx\n ",
        info->ac_strt, info->ac_cunt);

         /* Start Write Sector in block */
    info->secnm = (unsigned short)(info->ac_strt % info->SectorPerBlock);
         /* Block Allocation Table Point */
    info->ac_strt = info->ac_strt / info->SectorPerBlock;
    info->ac_zone = (unsigned short)(info->ac_strt / MAX_BLOCK_PER_ZONE_LOGICAL);
    info->ac_block = (unsigned short)(info->ac_strt % MAX_BLOCK_PER_ZONE_LOGICAL);

//    printk(" WriteZONE= %d, LogicalBlock= %04x, SectorInBlock= %02x, SecCount= %lx\n",
//      info->ac_zone, info->ac_block, info->secnm, info->ac_cunt );

    if ( info->TableFlag[info->ac_zone] == 0xffff) {   /* Invalid */
        if ( MakeBlockTableZone(us, info->ac_zone) )  { /* Make Block Allocation Table */
            printk(" MakeBlockTableZone Error : \n");
                        /*
                         *   Must set Sense data (No media)
                         */

            srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
            return USB_STOR_TRANSPORT_ERROR;
        }
    }

    if ( info->ComputeGeomFlag == 0 ) {
        printk(" info->ComputeGeomFlag = 0 : \n");

                // IORS_MEDIA_ERROR
            srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
            return USB_STOR_TRANSPORT_ERROR;
    }

    if (info->ReadOnly) {
                // WRITE PROTECT
        srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
        return USB_STOR_TRANSPORT_ERROR;
    }

    if ( info->BadBlockCountPerZone[info->ac_zone] > info->LimitBadBlock[info->ac_zone] ) {
        printk("BadBlock= %d, LimitBad= %d\n",
                info->BadBlockCountPerZone[info->ac_zone],
                info->LimitBadBlock[info->ac_zone]);
        printk(" badbK > LimitBad Error :\n\n");
        srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
        return USB_STOR_TRANSPORT_ERROR;
    }

    if (info->pagesize == Page512M )
        WriteBufferSize = info->SectorPerBlock * (512 + 64 );
    else
        WriteBufferSize = info->SectorPerBlock * (512 + 64 + 64);

    if ((info->MyMediaBuf
            = kmalloc( WriteBufferSize, GFP_KERNEL)) == NULL) {
        printk("sma03_Write: can't allocate MyMediaBuf size=%lx\n",
                WriteBufferSize);
        return USB_STOR_TRANSPORT_ERROR;
    }

    MediaBufPtr = (PBLOCKREAD_MEMORYMAP512)info->MyMediaBuf;
    w_count = 0;
    remain_count = info->ac_cunt;
    do {
        info->ac_zone
         = (unsigned short)(info->ac_strt / MAX_BLOCK_PER_ZONE_LOGICAL);
        info->ac_block
         = (unsigned short)(info->ac_strt % MAX_BLOCK_PER_ZONE_LOGICAL);
         /* Invalid */
        if ( info->TableFlag[info->ac_zone] == 0xffff ) {
            if ( MakeBlockTableZone( us, info->ac_zone ) ) {
                printk(" MakeBlockTableZone() Error : \n");
                kfree(info->MyMediaBuf);
                srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                return USB_STOR_TRANSPORT_ERROR;
            }
        }

        if ( info->BlockTable[info->ac_zone][info->ac_block] != 0xffff ) {
            US_DEBUGP("ENTRY SECNM = 0x%x\n", info->secnm);

            oldBlock = info->BlockTable[info->ac_zone][info->ac_block];
            erase_flag = TRUE;
            US_DEBUGP("ENTRY OldBlock = 0x%x\n", oldBlock);

            if ( info->secnm != 0 ) {
                RW_SectorCount = info->secnm;
                bstatus = SetOrginalData( us, MediaBufPtr, 0, RW_SectorCount );
                if ( bstatus ) {
                    printk(" return SectorWrite ERR SetOrginalData\n" );
                    kfree(info->MyMediaBuf);
                    srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                    return USB_STOR_TRANSPORT_ERROR;
                } else if ( bstatus == SMA_ECC_MARKED_F0 ) {
                    erase_flag = FALSE;
                }
                MediaBufPtr += RW_SectorCount;
            }

            if ( ((unsigned long)(info->SectorPerBlock - info->secnm))
                         > remain_count ) {
                RW_SectorCount = (unsigned short)remain_count;
//  printk("SetWriteNewData...: start: %p %p\n", MediaBufPtr, us->extra);
//  printk(" #0: DataArea: %p\n",  MediaBufPtr->DataArea);
              SetWriteNewData( us, MediaBufPtr, RW_SectorCount );
//  printk("SetWriteNewData...: end\n");
                MediaBufPtr += RW_SectorCount;
                w_count = w_count + remain_count;

                RW_SectorCount = info->SectorPerBlock
                                 - (info->secnm + (unsigned short)remain_count) ;
                bstatus = SetOrginalData( us, MediaBufPtr,
                                 (unsigned short)(info->secnm + remain_count),
                                 RW_SectorCount );
                if ( bstatus ) {
                    printk(" return SectorWrite ERR SetOrginalData\n");
                    kfree(info->MyMediaBuf);
                    srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                    return USB_STOR_TRANSPORT_ERROR;
                } else if ( bstatus == SMA_ECC_MARKED_F0 ) {
                    erase_flag = FALSE;
                }
                MediaBufPtr += RW_SectorCount;
            } else {
                RW_SectorCount = info->SectorPerBlock - info->secnm;
                SetWriteNewData(us, MediaBufPtr, RW_SectorCount );
                MediaBufPtr += RW_SectorCount;
                w_count = w_count + (info->SectorPerBlock - info->secnm);
            }
        } else {
            US_DEBUGP("Unused Block  secnm = 0x%x\n", info->secnm);
            erase_flag = FALSE;
            if ( info->secnm != 0 ) {
                RW_SectorCount = info->secnm;
                SetRedundantData( us, MediaBufPtr, 0, RW_SectorCount );
                MediaBufPtr += RW_SectorCount;
            }

            if ( ((unsigned long)(info->SectorPerBlock - info->secnm))
                         > remain_count) {
                RW_SectorCount = (unsigned short)remain_count;
                SetWriteNewData( us, MediaBufPtr, RW_SectorCount );
                MediaBufPtr += RW_SectorCount;

                w_count = w_count + remain_count;

                RW_SectorCount = info->SectorPerBlock
                                 - (info->secnm + (unsigned short)remain_count);
                SetRedundantData( us, MediaBufPtr,
                        (unsigned short)(info->secnm + (unsigned short)remain_count),
                        RW_SectorCount );
                MediaBufPtr += RW_SectorCount;
            } else {
                RW_SectorCount = info->SectorPerBlock - info->secnm;
                SetWriteNewData( us, MediaBufPtr, RW_SectorCount );
                MediaBufPtr += RW_SectorCount;
                w_count = w_count + (info->SectorPerBlock - info->secnm);
            }
        }

        if ( ( info->FWInfo.CodeLevel & CLEVEL_WRITEBLOCK_SUPPORT )
                    == CLEVEL_WRITEBLOCK_SUPPORT ) {
            if ( info->pagesize == Page256M ) {
                ConvertMediaBuffer512to256( us, info->SectorPerBlock );
            }
        }

        WriteFailCounter = 0;
        VerifyFailCounter = 0;
        UsbOutFailCounter = 0;

        while ( 1 ) {
            int   VerifyCheckStatus;

            VerifyFailCounter = 0;
            UsbOutFailCounter = 0;

            bstatus = CheckWriteEntryTable( us );
            if ( bstatus ) {   // error!!
                kfree(info->MyMediaBuf);
                srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                return USB_STOR_TRANSPORT_ERROR;
            }

            NewBlock = info->WriteEntryTableList[info->ac_zone][info->WritePoint[info->ac_zone]];
            US_DEBUGP(" Write PhysicalBlock = 0x%x\n", NewBlock );

            info->WriteEntryTableList[info->ac_zone][info->WritePoint[info->ac_zone]] = 0xffff;
            info->WritePoint[info->ac_zone]++ ;
            if ( info->WritePoint[info->ac_zone] >= info->tblsz[info->ac_zone] )
                info->WritePoint[info->ac_zone] = 0;

WRITE_VERIFY_RETRY0:
            VerifyCheckStatus = SMA_GOOD;

/*
            if ( ( info->FWInfo.CodeLevel & CLEVEL_WRITEBLOCK_SUPPORT )
                    == CLEVEL_WRITEBLOCK_SUPPORT ) {
                bstatus = new_BlockWrite( us, NewBlock,
                                info->SectorPerBlock, &MediaStatus );
            } else {
                bstatus = old_BlockWrite( us, NewBlock,
                                info->SectorPerBlock, &MediaStatus );
            }
*/
            bstatus = info->write_block(us, NewBlock,
                                 info->SectorPerBlock, &MediaStatus);
            if ( bstatus == SMA_ERROR_TRANSFER_WRITE_BOMB ) {
                bstatus = EraseForWriteRetry( us, NewBlock, &MediaStatus );
                if ( bstatus != SMA_GOOD ) {
                        goto WRITE_CHECK_STATUS;
                } else {
                        goto WRITE_VERIFY_RETRY0;
                }

                UsbOutFailCounter ++;
                if ( UsbOutFailCounter > 5 ) {
                    printk(" USB TRANSFER (MediaWrite)  BOMBER!!  [Catastrophe!!!]\n");
                    kfree(info->MyMediaBuf);
                    srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                    return USB_STOR_TRANSPORT_ERROR;
                }
                goto WRITE_VERIFY_RETRY0;
            } else if ( bstatus == SMA_GOOD ) {
                if ( ( info->WriteVerify )
                        && ( ( info->FWInfo.CodeLevel
                                & (CLEVEL_READBLOCK_SUPPORT | CLEVEL_READBLOCK_SUPPORT) )
                        == (CLEVEL_READBLOCK_SUPPORT | CLEVEL_READBLOCK_SUPPORT)  ) ) {
                    VerifyCheckStatus = VerifySub( us,
                                (PBLOCKREAD_MEMORYMAP512)info->MyMediaBuf,
                                NewBlock,
                                info->SectorPerBlock );
                    if ( VerifyCheckStatus != SMA_GOOD ) {
                        VerifyFailCounter ++;
                        if ( VerifyFailCounter > 3 ) {
                            printk(" [ERROR!!] WriteVerify : RetryOver!! : ERROR *** \n" );
                            goto WRITE_VERIFY_FAIL;
                        }
                        bstatus = EraseForWriteRetry( us, NewBlock, &MediaStatus );
                        if ( bstatus == SMA_GOOD ) {
                            goto WRITE_VERIFY_RETRY0;
                        } else {
                            printk(" EraseForWriteRetry : EraseBlock Fail\n" );
                            goto WRITE_CHECK_STATUS;
                        }

                    }
                }
                break;  //+++ EXIT while(1) : for WRITE_BLOCK RETRY
            }

WRITE_CHECK_STATUS:
            if ( bstatus == SMA_WRITE_PROTECTED ) {
                // WriteProtect
                printk(" WriteFail MEDIA BUSY!!! : Write Protrcted*** \n" );
                kfree(info->MyMediaBuf);
                srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                return USB_STOR_TRANSPORT_ERROR;
            }
            if ( !( MediaStatus & SM_STATUS_READY ) ) {
                 // MEDIA BUSY
                 //  NotReady
                printk(" WriteFail MEDIA BUSY!!! *** \n" );
                kfree(info->MyMediaBuf);
                srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
                return USB_STOR_TRANSPORT_ERROR;
            }

WRITE_VERIFY_FAIL:
            if ( ( MediaStatus & SM_STATUS_FAIL )
                    || ( VerifyCheckStatus == SMA_ERROR ) ) {
                BadBlockSet( us, NewBlock );
            }

            WriteFailCounter ++;
            if ( WriteFailCounter > 3 ) {
                printk(" WriteFail or VerifyFail RENZOKU HASSEI!!! *** \n" );
                kfree(info->MyMediaBuf);
                srb->result = CHECK_CONDITION << 1;
                return USB_STOR_TRANSPORT_ERROR;
            }
        }  //+++ while(1) : for WRITE_BLOCK RETRY END

        if ( erase_flag == TRUE ) {
            bstatus = EraseBlock( us, oldBlock );
            if ( bstatus ) {
                printk(" sma03_Write : EraseFail \n" );
                kfree(info->MyMediaBuf);
                srb->result = CHECK_CONDITION << 1;
                return USB_STOR_TRANSPORT_ERROR;
            }
        }

        info->BlockTable[info->ac_zone][info->ac_block] = NewBlock;

        MediaBufPtr = (PBLOCKREAD_MEMORYMAP512)info->MyMediaBuf;

        info->secnm = 0;
        info->ac_strt ++ ;
        remain_count = info->ac_cunt - w_count;

    } while ( w_count < info->ac_cunt );

    kfree(info->MyMediaBuf);

    current_sense_data = noSense;
    srb->result = 0x0000;
//    printk(" return sma03_Write !!!\n" );

    return USB_STOR_TRANSPORT_GOOD;

} /* sma03_Write */

/*
 * SCSI Command Emulation for the RATOC SMA03
 *
 * Mode Sense (1A)
 */
static int
sma03_ModeSense(Scsi_Cmnd *srb, struct us_data *us)
{
    int     result;
    unsigned char buf[8];
    unsigned char *p;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    static unsigned char mode_page_01[4] = { // write-protected for now
        0x03, 0x00, 0x80, 0x00
    };


    US_DEBUGP("sma03_ModeSense: Page: %02X\n", (srb->cmnd[2] & 0x3F) );

    // Read-write error recovery page: there needs to
    // be a check for write-protect here
    buf[0] = srb->cmnd[2] & 0x3F;
    if ( (buf[0] != 0x01) && (buf[0] != 0x3F) ) {
        return USB_STOR_TRANSPORT_ERROR;
    }
    if ( srb->request_bufflen < 4 )
        return USB_STOR_TRANSPORT_ERROR;

    memcpy((p = srb->request_buffer), mode_page_01, sizeof(mode_page_01));

    current_sense_data = NULL;
    if ( (result = sma03_MediaProtect(us))
        || (info->ReadOnly == 1) ) /* Write protected */
        p[2] |= 0x80;  /* on */
    else
        p[2] &= ~0x80; /* off */

    US_DEBUGP("sma03_ModeSense: MediaProtect: %02X "
                "write-protect: %02X\n", result, p[2] );

    return USB_STOR_TRANSPORT_GOOD;

}  /* sma03_ModeSense */

/*
 * SCSI Command Emulation for the RATOC SMA03
 *
 * Prevent/Allow Medium Removal (1E)
 */
static int
sma03_PreventAllowMediumRemoval(Scsi_Cmnd *srb, struct us_data *us)
{

    US_DEBUGP("sma03_PreventAllowMediumRemoval: %02X\n", srb->cmnd[4]);

    current_sense_data = illegalRequest;
    memcpy(srb->sense_buffer, current_sense_data, 18);
    srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
    return USB_STOR_TRANSPORT_GOOD;

}  /* sma03_PreventAllowMediumRemoval */

/*
 * SCSI Command Emulation for the RATOC SMA03
 *
 * Read Capacity (25)
 */
static int
sma03_ReadCapacity(Scsi_Cmnd *srb, struct us_data *us)
{
    unsigned char *ptr;
    unsigned long capacity;
    struct sma03_card_info *info = (struct sma03_card_info *)(us->extra);

    sma03_GetFwInfo(us);

    if ( !(capacity = sma03_get_capacity(us)) )  {
            current_sense_data = mediumNotPresent;
            memcpy(srb->sense_buffer, current_sense_data, 18);
            info->ComputeGeomFlag = 0;
            srb->result = (CHECK_CONDITION << 1) | (DID_ERROR << 16);
            return USB_STOR_TRANSPORT_FAILED;
    }

    info->capacity = capacity;
    for (info->pageshift=1;
            (info->pagesize>>info->pageshift);
            info->pageshift++);
        info->pageshift--;
    for (info->blockshift=1;
            (info->blocksize>>info->blockshift);
            info->blockshift++);
        info->blockshift--;
    info->blockmask = (1<<info->blockshift)-1;

    // Last page in the card

    capacity /= info->pagesize;
    capacity--;

    ptr = (unsigned char *)srb->request_buffer;

    ptr[0] = MSB_of(capacity>>16);
    ptr[1] = LSB_of(capacity>>16);
    ptr[2] = MSB_of(capacity&0xFFFF);
    ptr[3] = LSB_of(capacity&0xFFFF);

    // The page size

    ptr[4] = MSB_of(info->pagesize>>16);
    ptr[5] = LSB_of(info->pagesize>>16);
    ptr[6] = MSB_of(info->pagesize&0xFFFF);
    ptr[7] = LSB_of(info->pagesize&0xFFFF);

    US_DEBUGP("PageSize: %d, PageShift: %d\n",
             info->pagesize, info->pageshift);
    US_DEBUGP("BlockSize: %d, BlockShift: %d\n",
             info->blocksize, info->blockshift);

    if (info->pagesize == Page256M) {
            info->read_sector = sma03_ReadDoublePages;
            info->write_sector = sma03_WriteDoublePages;
    } else {
            info->read_sector = sma03_ReadSinglePage;
            info->write_sector = sma03_WriteSinglePage;
    }

    if (sma03_read_map(us)) {
        printk("sam03_read_map: Error\n");
        return USB_STOR_TRANSPORT_ERROR;
    }

    if (info->FWInfo.Reserved[0] == 0x01) {
        info->read_block = sma03_BlockRead;
        info->write_block = new_BlockWrite;
        info->erase_block = sma03New_EraseBlock;
    } else {
        info->read_block = sma03_ReadSectors;
        info->write_block = old_BlockWrite;
        info->erase_block = sma03Old_EraseBlock;
    }

    info->ComputeGeomFlag = 1;
    info->WriteVerify = 0;

    current_sense_data = noSense;
//    printk("ReadCapacity :end\n");

    return USB_STOR_TRANSPORT_GOOD;
}

/*
 * Transport for the RATOC SMA03
 */
int
sma03_transport(Scsi_Cmnd *srb, struct us_data *us)
{
    int result;
    int i;
    char string[64];
    unsigned char inquiry_response[36] = {
        0x00, 0x80, 0x00, 0x02, 0x1F, 0x00, 0x00, 0x00
    };

    if (!us->extra) {
        us->extra = kmalloc(
            sizeof(struct sma03_card_info), GFP_KERNEL);
        if (!us->extra)
            return USB_STOR_TRANSPORT_ERROR;
        memset(us->extra, 0, sizeof(struct sma03_card_info));
        us->extra_destructor = sma03_card_info_destructor;
    }

    US_DEBUGP("Cmd:%02X ", srb->cmnd[0]);

    /* Dummy up a response for INQUIRY since sma03 doesn't
       respond to INQUIRY commands */
    if (srb->cmnd[0] == INQUIRY) {
        memset(inquiry_response+8, 0, 28);
        fill_inquiry_response(us, inquiry_response, 36);
        return USB_STOR_TRANSPORT_GOOD;
    }

    if (srb->cmnd[0] == READ_CAPACITY) {
        return sma03_ReadCapacity(srb, us);
    }

    if (srb->cmnd[0] == MODE_SENSE) {
        return sma03_ModeSense(srb, us);
    }

    if (srb->cmnd[0] == ALLOW_MEDIUM_REMOVAL) {
        return sma03_PreventAllowMediumRemoval(srb, us);
    }

    if (srb->cmnd[0] == REQUEST_SENSE) {
        return sma03_RequestSense(srb, us);
    }

    if (srb->cmnd[0] == READ_10) {
        return sma03_Read(srb, us, 0); /* 1:ReadVerify-on 0:off */

    }

    if (srb->cmnd[0] == WRITE_10) {
        return sma03_Write(srb, us, 0); /* 1:ReadVerify-on 0:off */

    }

    // Pass TEST_UNIT_READY and REQUEST_SENSE through

    if (srb->cmnd[0] != TEST_UNIT_READY &&
            srb->cmnd[0] != REQUEST_SENSE) {
        printk("Unsupported Cmd:%02X ", srb->cmnd[0]);
        return USB_STOR_TRANSPORT_ERROR; // FIXME: sense buffer?
    }

	for (; srb->cmd_len<12; srb->cmd_len++)
		srb->cmnd[srb->cmd_len] = 0;

	srb->cmnd[1] = 0x20;

	string[0] = 0;
	for (i=0; i<12; i++)
	  sprintf(string+strlen(string), "%02X ", srb->cmnd[i]);

	US_DEBUGP("sma03: Send control for command %s\n",
		string);

        if ( srb->cmnd[0] == TEST_UNIT_READY ) {
              result = sma03_TestUnitReady(srb, us);
              return result;
        }
	if (srb->request_bufflen == 0)
		return USB_STOR_TRANSPORT_GOOD;

	if (srb->sc_data_direction == SCSI_DATA_WRITE ||
	    srb->sc_data_direction == SCSI_DATA_READ) {

		US_DEBUGP("sma03: %s %d bytes\n",
			srb->sc_data_direction==SCSI_DATA_WRITE ?
			  "sending" : "receiving",
			srb->request_bufflen);
/*                result = sma03_rw_sector(srb, us);
		return result;
*/
    }

    return USB_STOR_TRANSPORT_GOOD;

}  /* sma03_transport */

/*
 * Ioctl
 */
sma03_ioctl()
{
}

